from typing import Any, Dict, List, Optional

from geolib.geometry.one import Point
from geolib.models import BaseDataClass
from geolib.models.internal import Bool

from .internal import CPT as InternalCPT
from .internal import ExcavationType, InterpretationType, Layer
from .internal import Profile as InternalProfile
from .internal import ReductionCoreResistanceEnum, TimeOrderType


class CPT(BaseDataClass):
    """TODO To be expanded/generalized in GEOLib+

    measure_data need to be provided as a list of dicts::

        [{
            "z": float,  # negative
            "qc": float,
        }]

    Note that `timeorder_type` will force all other
    CPTs in the same model to the same `timeorder_type`.
    """

    cptname: str
    groundlevel: float
    depthrange: float = 0.10  # minimum layer thickness
    measured_data: List[Dict[str, float]]

    excavation_type: ExcavationType = ExcavationType.BEFORE
    timeorder_type: TimeOrderType = TimeOrderType.CPT_EXCAVATION_INSTALL

    interpretation_model: InterpretationType = InterpretationType.NEN_RULE
    interpretation_model_stressdependent: Bool = Bool.FALSE

    # GEF related values
    pre_excavation: float = 987654321.000
    void_value_depth: float = 987654321.000
    void_value_cone_resistance: float = 987654321.000
    void_value_pore_water_pressure: float = 987654321.000
    void_value_sleeve_friction: float = 987654321.000
    void_value_friction_number: float = 987654321.000
    void_value_equivalent_electronic_qc: float = 987000000.000000

    def to_profile(self, method, **kwargs) -> "Profile":
        """TODO GEOLib+ Interpret CPT by `method` to generate a Profile."""
        pass

    def _to_internal(self) -> InternalCPT:
        kwargs = self.dict()
        kwargs["measured_data"] = {"data": kwargs["measured_data"]}
        return InternalCPT(**kwargs)


class Excavation(BaseDataClass):
    """Excavation class.

    Note that using an excavation level will
    override all previous set excavation levels in the same model."""

    reduction_of_core_resistance: ReductionCoreResistanceEnum = (
        ReductionCoreResistanceEnum.SAFE
    )
    excavation_level: float
    excavation_width_infinite: Bool = Bool.TRUE
    excavation_length_infinite: Bool = Bool.TRUE
    distance_edge_pile_to_excavation_boundary: float = 0.0  # only valid for BEGEMANN


class Profile(BaseDataClass):
    """Generic profile class.

    Layers need to be provided as a list of dicts::

        [{
            "name": str,
            "material": str,
            "top_level": float  # [m]
            "excess_pore_pressure_top": float = 0.0  # [kN/m3]
            "excess_pore_pressure_bottom": float = 0.0  # [kN/m3]
            "ocr_value": float = 1.0  # [-]
            "reduction_core_resistance": int = 0  # [%]
        },]

    """

    name: str
    location: Point
    phreatic_level: float
    pile_tip_level: float
    cpt: CPT
    excavation: Optional[Excavation] = None
    overconsolidation_ratio: float = 1.0
    top_of_positive_skin_friction: float = 0.0
    bottom_of_negative_skin_friction: float = 0.0
    expected_ground_level_settlement: float = 0.0
    placement_depth_of_foundation: float = 0.0
    concentration_value_frohlich: int = 3
    top_tension_zone: float = 0.0
    # Any, as combined int, float will become int
    layers: List[Dict[str, Any]]

    def _to_internal(self, matching_cpt) -> InternalProfile:
        kwargs = self.dict(exclude={"cpt", "excavation", "location"})

        kwargs["matching_cpt"] = matching_cpt

        if self.excavation is not None:
            kwargs.update(self.excavation.dict())

        kwargs["x_coordinate"] = self.location.x
        kwargs["y_coordinate"] = self.location.y

        kwargs["layers"] = [Layer(**d) for d in self.layers]

        return InternalProfile(**kwargs)



"""
Profile and Layer classes which are used by both D-Foundations and DSheetPiling.
D-Foundations often requires more parameters, which are unused for DSheetPiling.

"""

from typing import List

from geolib.geometry.one import Point
from geolib.models import BaseDataClass
from geolib.soils import Soil


class CPT(BaseDataClass):
    """Internal Placeholder CPT measurements as used in D-Foundations/D-Sheetpiling.

    It is here because input files can contain CPTs coupled to a Profile.

    .. todo::
        Add Friction and other parameters?
    """

    z: List[float]
    qc: List[float]


class CPTRule(BaseDataClass):
    """Placeholder class for CPT interpretation options."""


class ProfileLayer(BaseDataClass):
    """Layer in a Profile. Note that we add the vertical location of
    the layer with a Point class.

    .. todo::
        * Will the ocr be added to the Soil class?
        * Decouple pore_pressures into special Class/Line
    """

    top: Point
    material: Soil
    pore_pressure_top: float  # absolute (needs calculation to ad.)
    pore_pressure_bottom: float  # absolute (needs calculation to ad.)
    ocr: float
    reduction_cone_resistance: float


class Profile(BaseDataClass):
    """D-Foundations Profile.

    .. todo::
        * Determine extra fields related to Excavations

    """

    label: str
    layers: List[ProfileLayer]
    phreatic_level: float
    pile_tip_level: float
    overconsolidation_ratio: float
    top_positive_skin_friction: float
    bottom_positive_skin_friction: float
    expected_ground_level_settlement: float
    min_layer_thickness: float
    cpt: CPT
    cpt_rule: CPTRule



from .base_model_structure import BaseDataClass, BaseModelStructure  # isort:skip
from .base_model import BaseModel, BaseModelList
from .dfoundations import DFoundationsModel
from .dsettlement import DSettlementModel
from .dsheetpiling import DSheetPilingModel
from .dstability import DStabilityModel
from .meta import MetaData
from .model_enums import Color
from .validators import BaseValidator



import logging
from enum import Enum, IntEnum
from inspect import cleandoc
from math import isclose
from operator import attrgetter
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from pydantic.types import PositiveInt, confloat, conint, conlist, constr

from geolib.geometry.one import Point
from geolib.models import BaseDataClass
from geolib.models.base_model_structure import BaseModelStructure
from geolib.models.dseries_parser import (
    DSerieListStructure,
    DSerieMatrixStructure,
    DSerieOldTableStructure,
    DSeriesInlineMappedProperties,
    DSeriesMatrixTreeStructureCollection,
    DSeriesNoParseSubStructure,
    DSeriesRepeatedGroupedProperties,
    DSeriesStructure,
    DSeriesStructureCollection,
    DSeriesTableStructure,
    DSeriesTreeStructure,
    DSeriesTreeStructureCollection,
    DSeriesUnmappedNameProperties,
    DSerieVersion,
)
from geolib.models.dsettlement.dsettlement_structures import (
    ComplexVerticalSubstructure,
    DSerieRepeatedTableStructure,
)
from geolib.models.dsettlement.internal_soil import SoilInternal
from geolib.models.dsettlement.probabilistic_calculation_types import (
    ProbabilisticCalculationType,
)
from geolib.models.internal import Bool
from geolib.models.utils import get_required_class_field
from geolib.soils import DistributionType, HorizontalBehaviourType, Soil
from geolib.soils import StorageTypes as StorageTypes_external
from geolib.utils import make_newline_validator

from .drain_types import DrainGridType, DrainSchedule, DrainType

logger = logging.getLogger(__name__)


TOLERANCE = 1e-10
ZERO_ITEMS = "    0 = number of items"
REQ_RUN_LINES = 2


class DSeriePoint(DSeriesTreeStructure):
    """
    Converting points from geolib co-ordinate system to the d-settlement UI
    Internal setting of the geometry differs from API.
    Here co-ordinate system is the same as in the the d-settlement UI.
    So axis z of the geolib needs to be modified to y axis, which represents the depth.
    """

    id: PositiveInt
    X: float
    Y: float
    Z: float

    tolerance: float = TOLERANCE

    @classmethod
    def get_structure_required_fields(cls) -> List[Tuple[str, Type]]:
        point_required_field_names = [
            required_field[0] for required_field in get_required_class_field(cls)
        ]
        return [
            field_tuple
            for field_tuple in super().get_structure_required_fields()
            if field_tuple[0] in point_required_field_names
        ]

    @classmethod
    def from_point(cls, p: Point):
        return cls(id=1, X=p.x, Y=p.z, Z=p.y)

    def __eq__(self, other: object):
        if isinstance(other, DSeriePoint):
            return (
                isclose(self.X, other.X, abs_tol=self.tolerance)
                and isclose(self.Y, other.Y, abs_tol=self.tolerance)
                and isclose(self.Z, other.Z, abs_tol=self.tolerance)
            )
        return NotImplemented


class SoilCollection(DSeriesStructureCollection):
    soil: List[SoilInternal] = []

    def add_soil_if_unique(self, soil, tolerance=TOLERANCE) -> None:
        for added_soil in self.soil:
            if soil.name == added_soil.name:
                raise NameError(f"Soil with name {soil.name} already exists.")
        self.soil.append(soil)


class Version(DSerieVersion):
    soil: int = 1005
    geometry: int = 1000
    d__settlement: int = 1007


class Points(DSeriesMatrixTreeStructureCollection):
    """Representation of [POINTS] group."""

    points: List[DSeriePoint] = []

    def add_point_if_unique(self, point: DSeriePoint, tolerance=TOLERANCE) -> DSeriePoint:
        if point in self.points:
            point = self.points[self.points.index(point)]
        else:
            point = self.add_point(point)
        return point

    def add_point(self, point: DSeriePoint) -> DSeriePoint:
        new_point_id = max(self.points, key=attrgetter("id")).id + 1 if self.points else 1
        point.id = new_point_id
        self.points.append(point)
        return point

    def __getitem__(self, point_id: int) -> Point:
        for point in self.points:
            if point.id == point_id:
                return point
        raise KeyError(f"Can't find point with id {point_id}.")


class Curve(DSeriesTreeStructure):
    """Curve is a Line consisting of two points (by reference)."""

    id: PositiveInt = 1
    points: conlist(int, min_items=2, max_items=2)

    def __eq__(self, other: object) -> bool:
        """
        Check if points in two lines are equal.

        Args:
            other (Line): an other line

        Returns: boolean

        """
        if not isinstance(other, Curve):
            return NotImplemented
        else:
            return self.points == other.points


class Curves(DSeriesTreeStructureCollection):
    """Representation of [CURVES] group."""

    curves: List[Curve] = []

    def create_curve(self, a: DSeriePoint, b: DSeriePoint):
        assert a.id != b.id
        curve = Curve(points=[a.id, b.id])
        if curve in self.curves:
            return self.curves[self.curves.index(curve)]
        else:
            new_curve_id = (
                max(self.curves, key=attrgetter("id")).id + 1
                if len(self.curves) > 0
                else 1
            )
            curve.id = new_curve_id
            self.curves.append(curve)

        return curve

    def create_curves(self, sorted_points: List[DSeriePoint]) -> List[Curve]:

        new_curves = [
            self.create_curve(sorted_points[i], sorted_points[i + 1])
            for i in range(len(sorted_points) - 1)
        ]
        return new_curves

    def __getitem__(self, curve_id: int) -> Curve:
        for curve in self.curves:
            if curve.id == curve_id:
                return curve
        raise KeyError(f"Can't find curve with id {curve_id}.")


class Boundary(DSeriesTreeStructure):
    id: conint(ge=0) = 0
    curves: List[int] = []

    def __eq__(self, other: object) -> bool:
        """
        Check if curves in two Boundary are equal.

        Arguments:
            other (Boundary): an other Boundary

        Returns: boolean

        """
        if not isinstance(other, Boundary):
            return NotImplemented
        else:
            return self.curves == other.curves


class Boundaries(DSeriesTreeStructureCollection):
    """Representation of [BOUNDARIES] group."""

    boundaries: List[Boundary] = []

    def create_boundary(self, curves: List[Curve]) -> Boundary:
        boundary = Boundary(curves=[curve.id for curve in curves])

        # Return the id of existing one
        if boundary in self.boundaries:
            return self.boundaries[self.boundaries.index(boundary)]
        else:
            new_boundary_id = (
                max(self.boundaries, key=attrgetter("id")).id + 1
                if len(self.boundaries) > 0
                else 0
            )
            boundary.id = new_boundary_id
            self.boundaries.append(boundary)

        return boundary

    def __getitem__(self, boundary_id: int) -> Boundary:
        for boundary in self.boundaries:
            if boundary.id == boundary_id:
                return boundary
        raise KeyError(f"Can't find boundary with id {boundary_id}.")


class Layer(DSeriesTreeStructure):
    id: PositiveInt = 1
    material: str = ""  # reference to soil name
    piezo_top: conint(ge=0, le=99) = 0  # reference to head line
    piezo_bottom: conint(ge=0, le=99) = 0  # reference to head line
    boundary_top: int  # reference to boundary
    boundary_bottom: int  # reference to boundary

    def __eq__(self, other: object) -> bool:
        """
        Check if curves in two Boundary are equal.

        Arguments:
            other (Boundary): an other Boundary

        Returns: boolean

        """
        if not isinstance(other, Layer):
            return NotImplemented
        else:
            return (
                self.boundary_top == other.boundary_top
                and self.boundary_bottom == other.boundary_bottom
            )


class Layers(DSeriesTreeStructureCollection):
    """Representation of [LAYERS] group."""

    layers: List[Layer] = []

    def add_layer(self, layer: Layer):

        if layer in self.layers:
            existing_layer = self.layers[self.layers.index(layer)]
            logger.warning(
                f"It's not possible to replace existing layers: {existing_layer}"
            )
            return existing_layer
        else:
            new_layer_id = (
                max(self.layers, key=attrgetter("id")).id + 1
                if len(self.layers) > 0
                else 1
            )
            layer.id = new_layer_id
            self.layers.append(layer)

        return layer

    def __getitem__(self, layer_id: int) -> Layer:
        for layer in self.layers:
            if layer.id == layer_id:
                return layer
        raise KeyError(f"Can't find layer with id {layer_id}.")


class PiezoLine(DSeriesTreeStructure):
    id: PositiveInt = 1
    curves: List[int] = []

    def __eq__(self, other: object):
        if isinstance(other, PiezoLine):
            return self.curves == other.curves
        return NotImplemented


class PiezoLines(DSeriesTreeStructureCollection):
    """Representation of [PIEZO LINES] group."""

    piezolines: List[PiezoLine] = []

    def create_piezoline(self, curves: List[Curve]) -> PiezoLine:
        piezoline = PiezoLine(curves=[curve.id for curve in curves])

        if piezoline in self.piezolines:
            return self.piezolines[self.piezolines.index(piezoline)]
        else:
            new_piezoline_id = (
                max(self.piezolines, key=attrgetter("id")).id + 1
                if len(self.piezolines) > 0
                else 1
            )
            piezoline.id = new_piezoline_id
            self.piezolines.append(piezoline)
        return piezoline

    def __getitem__(self, piezo_line_id: int) -> PiezoLine:
        for piezo_line in self.piezolines:
            if piezo_line.id == piezo_line_id:
                return piezo_line
        raise KeyError(f"Can't find headline with id {piezo_line_id}.")


class UseProbabilisticDefaultsBoundaries(DSerieListStructure):
    useprobabilisticdefaultsboundaries: List[Bool] = []

    def append_use_probabilistic_defaults_boundary(
        self, use_probabilistic_boundary: bool
    ):
        self.useprobabilisticdefaultsboundaries.append(
            Bool(int(use_probabilistic_boundary))
        )


class StdvBoundaries(DSerieListStructure):
    stdvboundaries: List[float] = []

    def append_stdv_boundary(self, stdv_boundary: float):
        self.stdvboundaries.append(stdv_boundary)


class DistributionBoundaries(DSerieListStructure):
    distributionboundaries: List[DistributionType] = []

    def append_distribution_boundary(self, distribution_boundary: DistributionType):
        if (distribution_boundary is DistributionType.Normal) or (
            distribution_boundary is DistributionType.Undefined
        ):
            self.distributionboundaries.append(distribution_boundary)
        else:
            raise ValueError(
                f"Enumeration member {repr(distribution_boundary)} is not supported for probabilistic boundary, please select Normal Distribution"
            )


class Accuracy(DSeriesUnmappedNameProperties):
    accuracy: confloat(ge=1e-10) = 1e-3


class PhreaticLine(DSeriesUnmappedNameProperties):
    phreatic_line: conint(ge=0, lt=99) = 0


class GeometryData(DSeriesStructure):
    """Representation of [GEOMETRY DATA] group."""

    accuracy: Accuracy = Accuracy()
    points: Points = Points()
    curves: Curves = Curves()
    boundaries: Boundaries = Boundaries()
    use_probabilistic_defaults_boundaries: UseProbabilisticDefaultsBoundaries = (
        UseProbabilisticDefaultsBoundaries()
    )
    stdv_boundaries: StdvBoundaries = StdvBoundaries()
    distribution_boundaries: DistributionBoundaries = DistributionBoundaries()
    piezo_lines: PiezoLines = PiezoLines()
    phreatic_line: PhreaticLine = PhreaticLine()
    world_co__ordinates: str = cleandoc(
        """
          0.000 - X world 1 -
          0.000 - Y world 1 -
          0.000 - X world 2 -
          0.000 - Y world 2 -
          """
    )
    layers: Layers = Layers()
    layerloads: str = ""

    def boundary_area_above_horizontal(self, boundary: Boundary, y: float = 0.0) -> float:
        """Area above horizontal line defined by y-coordinate."""
        area = 0.0

        for curve_id in boundary.curves:
            curve = self.get_curve(curve_id)
            points = [self.get_point(pid) for pid in curve.points]
            _y = (points[0].Y + points[1].Y) / 2 - y
            area += (points[1].X - points[0].X) * _y

        return area

    # Only do this before serializing, as it mutates
    # the ids of boundaries that we've already given
    # back to the user!
    def sort_boundaries(self):
        self.boundaries.boundaries.sort(
            key=lambda boundary: self.boundary_area_above_horizontal(boundary)
        )
        # sort probabilistic values
        self.sort_probabilistic_list_based_on_new_indexes()

        for i, boundary in enumerate(self.boundaries.boundaries):
            for layer in self.layers.layers:
                if layer.boundary_top == boundary.id:
                    layer.boundary_top = i
                if layer.boundary_bottom == boundary.id:
                    layer.boundary_bottom = i

            if boundary.id != i:
                logger.warning(
                    f"Boundary id {boundary.id} has changed to {i} in sorting process"
                )
            boundary.id = i

    def sort_probabilistic_list_based_on_new_indexes(self):
        new_indexes = self.create_list_of_indexes()
        self.use_probabilistic_defaults_boundaries.useprobabilisticdefaultsboundaries = self.sort_based_on_new_indexes(
            new_indexes,
            self.use_probabilistic_defaults_boundaries.useprobabilisticdefaultsboundaries,
        )
        self.stdv_boundaries.stdvboundaries = self.sort_based_on_new_indexes(
            new_indexes,
            self.stdv_boundaries.stdvboundaries,
        )
        self.distribution_boundaries.distributionboundaries = (
            self.sort_based_on_new_indexes(
                new_indexes,
                self.distribution_boundaries.distributionboundaries,
            )
        )

    def pre_process(self):
        self.sort_boundaries()
        self.sort_probabilistic_list_based_on_new_indexes()

    def create_list_of_indexes(self) -> List:
        return [boundary.id for boundary in self.boundaries.boundaries]

    def sort_based_on_new_indexes(self, new_indexes: List, unsorted_list: List) -> List:
        return [unsorted_list[i] for i in new_indexes]

    def get_point(self, point_id: int) -> Optional[DSeriePoint]:
        return self.points[point_id]

    def get_curve(self, curve_id: int) -> Optional[Curve]:
        return self.curves[curve_id]

    def get_boundary(self, boundary_id: int) -> Optional[Boundary]:
        return self.boundaries[boundary_id]

    def get_layer(self, layer_id: int) -> Optional[Layer]:
        return self.layers[layer_id]


class PointForLoad(BaseDataClass):
    """Different from DSeriePoint as it does not
    belong in [POINTS]."""

    X: float
    Y: float

    @classmethod
    def from_point(cls, p: Point):
        return cls(X=p.x, Y=p.z)


class NonUniformLoad(BaseDataClass):
    time: int = 0
    gammadry: float = 10
    gammawet: float = 10
    temporary: Bool = False
    endtime: int = 0
    points: List[PointForLoad]


class NonUniformLoads(DSeriesNoParseSubStructure):
    """Representation of [NON-UNIFORM LOADS] group."""

    loads: Dict[constr(min_length=1, max_length=25), NonUniformLoad] = {}

    def add_load(
        self, name: constr(min_length=1, max_length=25), load: NonUniformLoad
    ) -> Optional[ValueError]:
        if name in self.loads:
            raise ValueError(f"Load with name '{name}' already exists.")
        else:
            self.loads[name] = load
            self.loads = self.loads  # trigger validation
            return None


class WaterLoad(BaseDataClass):
    name: str = ""
    time: int = 0
    phreatic_line: int = 1
    headlines: List[List[int]] = [[]]


class WaterLoads(DSeriesNoParseSubStructure):
    """Representation of [WATER LOADS] group."""

    waterloads: List[WaterLoad] = []

    def add_waterload(
        self, name: str, time: int, phreatic_line: int, headlines: List[List[int]]
    ):
        self.waterloads.append(
            WaterLoad(
                name=name, time=time, phreatic_line=phreatic_line, headlines=headlines
            )
        )


class ResidualTimes(DSeriesNoParseSubStructure):
    """Representation of [RESIDUAL TIMES] group."""

    time_steps: List[conint(ge=0)] = []


class Verticals(DSeriesNoParseSubStructure):
    """Representation of [VERTICALS] group."""

    # total mesh is default value that is written in sli file but not read
    total_mesh: int = 100
    locations: List[DSeriePoint] = []


class TypeOtherLoads(IntEnum):
    Trapeziform = 0
    Circular = 1
    Rectangular = 2
    Uniform = 3
    Tank = 4


class LoadValuesTrapeziform(BaseDataClass):
    gamma: float = 0
    height: float = 0
    xl: float = 0
    xm: float = 0
    xr: float = 0
    Xp: float = 0
    Yp: float = 0


class LoadValuesCircular(BaseDataClass):
    weight: float = 0
    alpha: float = 0
    Xcp: float = 0
    Ycp: float = 0
    Zcp: float = 0
    R: float = 0.01


class LoadValuesRectangular(BaseDataClass):
    weight: float = 0
    alpha: float = 0
    Xcp: float = 0
    Ycp: float = 0
    Zcp: float = 0
    xwidth: float = 0.01
    zwidth: float = 0.01


class LoadValuesUniform(BaseDataClass):
    unit_weight: float = 0
    gamma: float = 0
    height: float = 0
    y_application: float = 0


class LoadValuesTank(BaseDataClass):
    wallweight: float = 0
    internalweight: float = 0
    alpha: float = 0
    Xcp: float = 0
    Ycp: float = 0
    Zcp: float = 0
    Rintern: float = 0.01
    dWall: float = 0.01


class OtherLoad(BaseDataClass):
    load_type: TypeOtherLoads
    time: int = 0
    load_values_trapeziform: Optional[LoadValuesTrapeziform]
    load_values_circular: Optional[LoadValuesCircular]
    load_values_rectangular: Optional[LoadValuesRectangular]
    load_values_uniform: Optional[LoadValuesUniform]
    load_values_tank: Optional[LoadValuesTank]


class OtherLoads(DSeriesNoParseSubStructure):
    """Representation of [OTHER LOADS] group."""

    loads: Dict[constr(min_length=1, max_length=25), OtherLoad] = {}

    def add_load(
        self, name: constr(min_length=1, max_length=25), load: OtherLoad
    ) -> Optional[ValueError]:
        if name in self.loads:
            raise ValueError(f"Load with name '{name}' already exists.")
        else:
            self.loads[name] = load
            return None


class Dimension(Enum):
    ONE_D = 0
    TWO_D = 1


class ConsolidationModel(Enum):
    DARCY = 0
    TERZAGHI = 1


class SoilModel(Enum):
    NEN_KOPPEJAN = 0
    NEN_BJERRUM = 1
    ISOTACHE = 2


class StrainType(Enum):
    LINEAR = 0
    NATURAL = 1


class Model(DSeriesNoParseSubStructure):
    dimension: Dimension = Dimension.TWO_D
    consolidation_model: ConsolidationModel = ConsolidationModel.DARCY
    soil_model: SoilModel = SoilModel.NEN_KOPPEJAN
    strain_type: StrainType = StrainType.LINEAR
    is_vertical_drains: Bool = Bool.FALSE
    is_fit_for_settlement_plate: Bool = Bool.FALSE
    is_probabilistic: Bool = Bool.FALSE
    is_horizontal_displacements: Bool = Bool.FALSE
    is_secondary_swelling: Bool = Bool.FALSE
    is_waspan: Bool = Bool.FALSE


class PreconPressureWithinLayer(Enum):
    CONSTANT_NO_CORRECTION = 0
    CONSTANT_CORRECTION_T0 = 1
    CONSTANT_CORRECTION_ALL_T = 2
    VARIABLE_NO_CORRECTION = 3
    VARIABLE_CORRECTION_T0 = 4
    VARIABLE_CORRECTION_ALL_T = 5


class DispersionConditionLayerBoundary(Enum):
    UNDRAINED = 0
    DRAINED = 1


class DispersionConditionsLayerBoundaries(DSeriesNoParseSubStructure):
    dispersion_conditions_layer_boundaries_top: DispersionConditionLayerBoundary = (
        DispersionConditionLayerBoundary.DRAINED
    )
    dispersion_conditions_layer_boundaries_bottom: DispersionConditionLayerBoundary = (
        DispersionConditionLayerBoundary.DRAINED
    )


class StressDistributionSoil(Enum):
    BUISMAN = 0
    BOUSSINESQ = 1


class StressDistributionLoads(Enum):
    NONE = 0
    SIMULATE = 1


class CalculationOptions(DSeriesNoParseSubStructure):
    precon_pressure_within_layer: PreconPressureWithinLayer = (
        PreconPressureWithinLayer.CONSTANT_NO_CORRECTION
    )
    is_imaginary_surface: Bool = Bool.FALSE
    imaginary_surface_layer: Optional[PositiveInt]
    is_submerging: Bool = Bool.FALSE
    use_end_time_for_fit: Bool = Bool.FALSE
    is_maintain_profile: Bool = Bool.FALSE
    maintain_profile_material_name: str = "Superelevation"
    maintain_profile_time: conint(ge=0, le=100000) = 0
    maintain_profile_gamma_dry: confloat(ge=-100, le=100) = 10
    maintain_profile_gamma_wet: confloat(ge=-100, le=100) = 10
    dispersion_conditions_layer_boundaries_top: DispersionConditionLayerBoundary = (
        DispersionConditionLayerBoundary.DRAINED
    )
    dispersion_conditions_layer_boundaries_bottom: DispersionConditionLayerBoundary = (
        DispersionConditionLayerBoundary.DRAINED
    )
    stress_distribution_soil: StressDistributionSoil = StressDistributionSoil.BUISMAN
    stress_distribution_loads: StressDistributionLoads = StressDistributionLoads.SIMULATE
    iteration_stop_criteria_submerging: confloat(ge=0.0, le=1.0) = 0.0
    iteration_stop_criteria_submerging_layer_height: confloat(ge=0, le=99.999) = 0
    maximum_iteration_steps_for_submerging: conint(ge=1, le=100) = 1
    iteration_stop_criteria_desired_profile: confloat(ge=0, le=1) = 0.1
    load_column_width_imaginary_surface: confloat(ge=0.05, le=10000) = 1
    load_column_width_non_uniform_loads: confloat(ge=0.05, le=10000) = 1
    load_column_width_trapeziform_loads: confloat(ge=0.05, le=10000) = 1
    end_of_consolidation: conint(ge=1, le=100000) = 100000
    number_of_subtime_steps: conint(ge=1, le=100) = 2
    reference_time: confloat(ge=0.001, le=1000000) = 1
    dissipation: Bool = Bool.FALSE
    x_coord_dissipation: float = 0.0
    use_fit_factors: Bool = Bool.FALSE
    x_coord_fit: float = 0.0
    is_predict_settlements_omitting_additional_load_steps: Bool = Bool.FALSE

    @classmethod
    def set_options(cls, **kwargs):
        cls_instance = cls(**kwargs)
        if cls_instance.is_imaginary_surface == Bool.FALSE:
            cls_instance.imaginary_surface_layer = None
        elif cls_instance.imaginary_surface_layer is None:
            cls_instance.imaginary_surface_layer = 1
        return cls_instance


class VerticalDrain(DSeriesNoParseSubStructure):
    # geometry of drain
    drain_type: DrainType = DrainType.STRIP
    range_from: confloat(ge=-10000000.000, le=10000000.000) = 0
    range_to: confloat(ge=-10000000.000, le=10000000.000) = 0
    bottom_position: confloat(ge=-10000000.000, le=10000000.000) = 0
    center_to_center: confloat(ge=0.001, le=1000) = 3
    width: confloat(ge=0.001, le=1000) = 0.1
    diameter: confloat(ge=0.001, le=1000) = 0.1
    thickness: confloat(ge=0.001, le=1000) = 0.003
    grid: DrainGridType = DrainGridType.UNDERDETERMINED
    # schedule
    schedule_type: DrainSchedule = DrainSchedule.OFF
    begin_time: confloat(ge=0, le=100000) = 0
    end_time: confloat(ge=0, le=100000) = 0
    under_pressure_for_strips_and_columns: confloat(ge=0, le=100000) = 35
    under_pressure_for_sand_wall: confloat(ge=0, le=100000) = 35
    start_of_drainage: confloat(ge=0, le=100000) = 0
    phreatic_level_in_drain: confloat(ge=-10000000.000, le=10000000.000) = 0
    water_head_during_dewatering: confloat(
        ge=-10000000.000, le=10000000.000
    ) = 0  # relevant for strip and column
    tube_pressure_during_dewatering: confloat(
        ge=-10000000.000, le=10000000.000
    ) = 0  # relevant for the sand wall
    time: List[confloat(ge=0, le=100000)] = []
    underpressure: List[confloat(ge=0, le=100000)] = []
    water_level: List[confloat(ge=-10000000.000, le=10000000.000)] = []


class InternalProbabilisticCalculationType(IntEnum):
    FOSMOrDeterministic = 0
    ProbabilityOfFailureFORM = 1
    BandWidthAndProbabilityOfFailureMonteCarlo = 2


class ProbabilisticData(DSeriesInlineMappedProperties):
    reliability_x_co__ordinate: float = 0
    residual_settlement: confloat(ge=0, le=1000) = 1
    maximum_drawings: conint(ge=0, le=999999999) = 100
    maximum_iterations: conint(ge=1, le=50) = 15
    reliability_type: InternalProbabilisticCalculationType = (
        InternalProbabilisticCalculationType.FOSMOrDeterministic
    )
    is_reliability_calculation: Bool = Bool.FALSE

    def set_probabilistic_data(
        self,
        point_of_vertical: Point,
        residual_settlement: float,
        maximum_number_of_samples: int,
        maximum_iterations: int,
        reliability_type: ProbabilisticCalculationType,
        is_reliability_calculation: bool,
    ):
        if (
            reliability_type == ProbabilisticCalculationType.SettlementsDeterministic
            and is_reliability_calculation
        ):
            raise ValueError(
                f"is_reliability_calculation is set to True but reliability type {repr(reliability_type)} is not probabilistic."
            )
        return ProbabilisticData(
            reliability_x_co__ordinate=point_of_vertical.x,
            residual_settlement=residual_settlement,
            maximum_drawings=maximum_number_of_samples,
            maximum_iterations=maximum_iterations,
            reliability_type=self.externaltointernalprobabilisticcalculationtype(
                reliability_type
            ),
            is_reliability_calculation=Bool(int(is_reliability_calculation)),
        )

    def externaltointernalprobabilisticcalculationtype(
        self, reliability_type: ProbabilisticCalculationType
    ) -> InternalProbabilisticCalculationType:
        map_values = {
            -1: InternalProbabilisticCalculationType.FOSMOrDeterministic,
            0: InternalProbabilisticCalculationType.FOSMOrDeterministic,
            1: InternalProbabilisticCalculationType.ProbabilityOfFailureFORM,
            2: InternalProbabilisticCalculationType.BandWidthAndProbabilityOfFailureMonteCarlo,
        }
        return map_values[reliability_type.value]


class DSettlementStructure(DSeriesStructure):
    """Representation of complete .sli file."""

    version: Version = Version()
    soil_collection: SoilCollection = SoilCollection()
    geometry_data: GeometryData = GeometryData()
    geometry_1d_data: Optional[str]
    run_identification: str = 2 * "\n"
    model: Union[Model, str] = Model()
    verticals: Union[Verticals, str] = Verticals()
    water: Union[float, str] = 9.81
    non__uniform_loads: Union[NonUniformLoads, str] = NonUniformLoads()
    water_loads: Union[WaterLoads, str] = WaterLoads()
    other_loads: Union[OtherLoads, str] = OtherLoads()
    calculation_options: Union[CalculationOptions, str] = CalculationOptions()
    residual_times: Union[ResidualTimes, str] = ResidualTimes()
    filter_band_width: str = cleandoc(
        """
        1 : Number of items
        0.05
        """
    )
    pore_pressure_meters: str = ZERO_ITEMS
    non__uniform_loads_pore_pressures: str = ZERO_ITEMS
    other_loads_pore_pressures: str = ZERO_ITEMS
    calculation_options_pore_pressures: str = cleandoc(
        """
        1 : Shear stress = TRUE
        1 : calculation method of lateral stress ratio (k0) = Nu
        """
    )
    vertical_drain: Union[VerticalDrain, str] = VerticalDrain()
    probabilistic_data: ProbabilisticData = ProbabilisticData()
    probabilistic_defaults: str = cleandoc(
        """
        ProbDefGamDryVar=0.05
        ProbDefGamWetVar=0.05
        ProbDefPOPVar=0.25
        ProbDefOCRVar=0.25
        ProbDefPcVar=0.25
        ProbDefPermeabilityVerVar=2.50
        ProbDefRatioHorVerPermeabilityCvVar=0.25
        ProbDefCvVar=0.50
        ProbDefCpVar=0.30
        ProbDefCp1Var=0.30
        ProbDefCsVar=0.30
        ProbDefCs1Var=0.30
        ProbDefApVar=0.30
        ProbDefASecVar=0.30
        ProbDefRRCrVar=0.25
        ProbDefCRCcVar=0.25
        ProbDefCaVar=0.25
        ProbDefPriCompIndexVar=0.25
        ProbDefSecCompIndexVar=0.25
        ProbDefSecCompRateVar=0.25
        ProbDefCpCor=0.01
        ProbDefCsCor=0.01
        ProbDefCs1Cor=0.01
        ProbDefApCor=0.01
        ProbDefASecCor=0.01
        ProbDefRRCrCor=0.01
        ProbDefCaCor=0.01
        ProbDefPriCompIndexCor=0.01
        ProbDefSecCompRateCor=0.01
        ProbDefGamDryDist=2
        ProbDefGamWetDist=2
        ProbDefPOPDist=2
        ProbDefOCRDist=2
        ProbDefPcDist=2
        ProbDefPermeabilityVerDist=2
        ProbDefRatioHorVerPermeabilityCvDist=2
        ProbDefCvDist=2
        ProbDefCpDist=2
        ProbDefCp1Dist=2
        ProbDefCsDist=2
        ProbDefCs1Dist=2
        ProbDefApDist=2
        ProbDefASecDist=2
        ProbDefRRCrDist=2
        ProbDefCRCcDist=2
        ProbDefCaDist=2
        ProbDefPriCompIndexDist=2
        ProbDefSecCompIndexDist=2
        ProbDefSecCompRateDist=2
        ProbDefLayerStd=0.10
        ProbDefLayerDist=0
        """
    )
    fit_options: str = cleandoc(
        """
        Fit Maximum Number of Iterations=5
        Fit Required Iteration Accuracy=0.0001000000
        Fit Required Correlation Coefficient=0.990
        """
    )
    fit_calculation: str = cleandoc(
        """
        Is Fit Calculation=0
        Fit Vertical Number=-1
        """
    )
    eps: str = cleandoc(
        """
        0.00 = Dry unit weight
        0.00 = Saturated unit weight
        0.00 = Load
        0.00 = Height above surface
        """
    )
    fit: str = ZERO_ITEMS

    # Custom validator
    _validate_run_identification = make_newline_validator(
        "run_identification", req_newlines=REQ_RUN_LINES
    )

    def validate_options(self):
        """
        Todo validation calculation options with geometry. I.e. validate x coordinate related options and layer related
            options
        Returns:

        """

    def check_x_in_vertical(self, point_of_vertical: Point):
        check_that_x_in_verticals = False
        for point in self.verticals.locations:
            if abs(point.X - point_of_vertical.x) <= point_of_vertical.tolerance:
                check_that_x_in_verticals = True
        if not (check_that_x_in_verticals):
            raise ValueError(
                f"The x-coordinate of point_of_vertical does not correspond to an existing vertical."
            )

    def get_headlines_for_layers(self):
        return [[l.piezo_top, l.piezo_bottom] for l in self.geometry_data.layers.layers]


class Stresses(DSeriesTableStructure):
    stresses: List[Dict[str, float]]


class KoppejanSettlements(DSeriesTableStructure):
    koppejansettlements: List[Dict[str, float]]


class Depths(DSerieListStructure):
    depths: List[float]


class Leakages(DSerieListStructure):
    leakages: List[float]


class DrainedLayers(DSerieListStructure):
    drainedlayers: List[int]


class TimeSettlementPerLoad(DSerieMatrixStructure):
    timesettlementperload: List[List[float]]


class HorizontalDisplacements(DSerieMatrixStructure):
    horizontaldisplacements: List[List[float]]


class TimeDependentData(DSerieRepeatedTableStructure):
    timedependentdata: Dict[float, List[Dict[str, float]]]


class Vertical(ComplexVerticalSubstructure):
    """Representation of [Vertical] group in sld file."""

    id: int
    x: float
    z: float
    time__settlement_per_load: Optional[TimeSettlementPerLoad]
    depths: Depths
    leakages: Optional[Leakages]
    drained_layers: Optional[DrainedLayers]
    stresses: Optional[Stresses]
    koppejan_settlement: Optional[KoppejanSettlements]
    time__dependent_data: List[TimeDependentData]
    elasticity: Optional[float]
    horizontal_displacements: Optional[HorizontalDisplacements]


class ResidualSettlements(DSerieOldTableStructure):
    # TODO LIst[Dict[str, float]] but can be empty which now gives a validation error
    residualsettlements: List[Dict[str, float]]


class Results(DSeriesRepeatedGroupedProperties):
    """Representation of [results] group in sld file."""

    verticals_count: int
    vertical: List[Vertical]
    residual_settlements: List[ResidualSettlements]
    amounts_of_loads: Optional[str]
    dissipation_in_layers: Optional[str]
    reliability_calculation_results: Optional[str]


class DSettlementOutputStructure(DSeriesStructure):
    """Representation of complete .sld file, inherting
    the structure of the .sli file as well."""

    results: Results
    input_data: DSettlementStructure



import logging
from enum import Enum, IntEnum
from inspect import cleandoc
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from pydantic.types import PositiveInt, confloat, conint, constr

from geolib.models import BaseDataClass
from geolib.models.base_model_structure import BaseModelStructure
from geolib.models.dseries_parser import (
    DSerieListStructure,
    DSerieMatrixStructure,
    DSerieOldTableStructure,
    DSeriesInlineMappedProperties,
    DSeriesInlineProperties,
    DSeriesNoParseSubStructure,
    DSeriesStructure,
    DSeriesStructureCollection,
    DSeriesTreeStructure,
    DSeriesTreeStructureCollection,
    DSerieVersion,
)
from geolib.models.internal import Bool
from geolib.utils import make_newline_validator

from .dfoundations_structures import (
    DFoundationsCPTCollectionWrapper,
    DFoundationsEnumStructure,
    DFoundationsInlineProperties,
    DFoundationsTableWrapper,
)
from .internal_soil import Soil

logger = logging.getLogger(__name__)


UNKNOWN = "Unknown"
REQ_RUN_LINES = 6


class PileType(IntEnum):
    PREFABRICATED_CONCRETE_PILE = 0
    CLOSED_ENDED_STEEL_PIPE_PILE = 1
    DRIVEN_CAST_IN_PLACE_PILE_TUBE_BACK_BY_DRIVING = 2
    DRIVEN_CAST_IN_PLACE_PILE_TUBE_BACK_BY_VIBRATION = 3
    TAPERED_TIMBER_PILE = 4
    STRAIGHT_TIMBER_PILE = 5
    SCREW_PILE_CAST_IN_PLACE_LOST_TIP = 6
    SCREW_PILE_CAST_IN_PLACE_WITH_GROUT = 7
    PREFABRICATED_SCREW_PILE_WITH_GROUT = 8
    PREFABRICATED_SCREW_PILE_WITHOUT_GROUT = 9
    STEEL_SECTION = 10
    CONTINUOUS_FLIGHT_AUGER_PILE = 11
    BORED_PILE_DRILLING_MUD_UNCASED_BOREHOLE = 12
    BORED_PILE_SHELLING_TECH_PERMANENT_CASING = 13
    OPEN_ENDED_STEEL_PIPE_PILE = 14
    MV_PILE = 15
    MICRO_PILE_DOUBLE_PIPE_EXTORTED = 16
    MICRO_PILE_DOUBLE_PIPE_NOT_EXTORTED = 17
    MICRO_PILE_SINGLE_PIPE_EXTORTED = 18
    MICRO_PILE_SINGLE_PIPE_NOT_EXTORTED = 19
    MICRO_PILE_ANCHOR_BORED = 20
    MICRO_PILE_ANCHOR_SCREWED = 21
    MICRO_PILE_VIBRATED = 22
    GROUTED_STEEL_PROFILE_WITH_FOOTPLATE = 23
    GROUTED_STEEL_PIPE_PILE_SCREWED = 24
    USER_DEFINED_VIBRATING = 25
    USER_DEFINED_LOW_VIBRATING = 26
    USER_DEFINED = 27


class PileTypeForClayLoamPeat(IntEnum):
    STANDARD = 0
    USER_DEFINED = 1


class LoadSettlementCurve(IntEnum):
    ONE = 0
    TWO = 1
    THREE = 2


class PileMaterial(IntEnum):
    CONCRETE = 0
    STEEL = 1
    TIMBER = 2
    USER_DEFINED = 3


class BearingPileSlipLayer(IntEnum):
    NONE = 0
    SYNTHETIC = 1
    BENTONITE = 2
    BITUMEN = 3
    USER_DEFINED = 4


class PileShape(IntEnum):
    ROUND_PILE = 0
    RECTANGULAR_PILE = 1
    ROUND_PILE_WITH_ENLARGED_BASE = 2
    RECTANGULAR_PILE_WITH_ENLARGED_BASE = 3
    ROUND_TAPERED_PILE = 4
    ROUND_HOLLOW_PILE_WITH_CLOSED_BASE = 5
    ROUND_PILE_WITH_LOST_TIP = 6
    ROUND_PILE_WITH_IN_SITU_FORMED_BASE = 7
    SECTION = 8
    ROUND_OPEN_ENDED_HOLLOW_PILE = 9
    H_SHAPED_PROFILE = 10
    USER_DEFINED = 11


class TypesBearingPiles(DSeriesNoParseSubStructure):
    pile_name: str = ""
    pile_type: PileType = PileType.PREFABRICATED_CONCRETE_PILE
    pile_type_for_execution_factor_sand_gravel: Optional[PileType]
    execution_factor_sand_gravel: Optional[confloat(ge=0, le=9)]
    pile_type_for_execution_factor_clay_loam_peat: Optional[PileTypeForClayLoamPeat]
    execution_factor_clay_loam_peat: Optional[confloat(ge=0, le=9)]
    pile_type_for_pile_class_factor: Optional[PileType]
    pile_class_factor: Optional[confloat(ge=0, le=9)]
    pile_type_for_load_settlement_curve: Optional[LoadSettlementCurve]
    material: Optional[PileMaterial]
    elasticity_modulus: Optional[confloat(ge=0, le=1e25)]
    slip_layer: BearingPileSlipLayer = BearingPileSlipLayer.NONE
    characteristic_adhesion: Optional[confloat(ge=0, le=1000)]
    shape: PileShape = PileShape.RECTANGULAR_PILE
    base_width: Optional[confloat(ge=0, le=100)]
    base_length: Optional[confloat(ge=0, le=100)]
    diameter: Optional[confloat(ge=0, le=100)]
    base_diameter: Optional[confloat(ge=0, le=100)]
    pile_diameter: Optional[confloat(ge=0, le=100)]
    base_height: Optional[confloat(ge=0, le=100)]
    base_width_v: Optional[confloat(ge=0, le=100)]
    base_length_v: Optional[confloat(ge=0, le=100)]
    shaft_width: Optional[confloat(ge=0, le=100)]
    shaft_length: Optional[confloat(ge=0, le=100)]
    increase_in_diameter: Optional[confloat(ge=0, le=100)]
    external_diameter: Optional[confloat(ge=0, le=100)]
    internal_diameter: Optional[confloat(ge=0, le=100)]
    height_h_shape: Optional[confloat(ge=0, le=100)]
    width_h_shape: Optional[confloat(ge=0, le=100)]
    thickness_web: Optional[confloat(ge=0, le=100)]
    thickness_flange: Optional[confloat(ge=0, le=100)]
    overrule_pile_tip_shape_factor: Bool = Bool.FALSE
    pile_tip_shape_factor: Optional[confloat(ge=0, le=10)]
    overrule_pile_tip_cross_section_factors: Bool = Bool.FALSE
    pile_tip_cross_section_factor: Optional[confloat(ge=0, le=10)]
    use_pre_2016: Bool = Bool.FALSE
    user_defined_pile_type_as_prefab: Bool = Bool.FALSE
    use_manual_reduction_for_qc: Bool = Bool.FALSE
    reduction_percentage_qc: confloat(ge=25, le=100) = 25
    is_user_defined: Bool = Bool.TRUE


class TypesTensionPiles(DSeriesNoParseSubStructure):
    pile_name: str = ""
    pile_type: PileType = PileType.PREFABRICATED_CONCRETE_PILE
    pile_type_for_execution_factor_sand_gravel: Optional[PileType]
    execution_factor_sand_gravel: Optional[confloat(ge=0, le=9)]
    pile_type_for_execution_factor_clay_loam_peat: Optional[PileTypeForClayLoamPeat]
    execution_factor_clay_loam_peat: Optional[confloat(ge=0, le=9)]
    material: Optional[PileMaterial]
    unit_weight_pile: Optional[confloat(ge=0, le=1000)]
    elasticity_modulus: Optional[confloat(ge=0, le=1e25)]
    shape: PileShape = PileShape.RECTANGULAR_PILE
    base_width: Optional[confloat(ge=0, le=100)]
    base_length: Optional[confloat(ge=0, le=100)]
    diameter: Optional[confloat(ge=0, le=100)]
    base_diameter: Optional[confloat(ge=0, le=100)]
    pile_diameter: Optional[confloat(ge=0, le=100)]
    base_height: Optional[confloat(ge=0, le=100)]
    base_width_v: Optional[confloat(ge=0, le=100)]
    base_length_v: Optional[confloat(ge=0, le=100)]
    shaft_width: Optional[confloat(ge=0, le=100)]
    shaft_length: Optional[confloat(ge=0, le=100)]
    increase_in_diameter: Optional[confloat(ge=0, le=100)]
    external_diameter: Optional[confloat(ge=0, le=100)]
    internal_diameter: Optional[confloat(ge=0, le=100)]
    height_h_shape: Optional[confloat(ge=0, le=100)]
    width_h_shape: Optional[confloat(ge=0, le=100)]
    thickness_web: Optional[confloat(ge=0, le=100)]
    thickness_flange: Optional[confloat(ge=0, le=100)]
    circumference: Optional[confloat(ge=0, le=100)]
    cross_section: Optional[confloat(ge=0, le=100)]
    is_user_defined: Bool = Bool.TRUE


class SoilCollection(DSeriesStructureCollection):
    soil: List[Soil] = Soil.default_soils()

    def add_soil_if_unique(self, soil) -> None:
        for added_soil in self.soil:
            if soil.name == added_soil.name:
                raise NameError(f"Soil with name {soil.name} already exists.")
        self.soil.append(soil)
        return soil

    def find_soil_id(self, key) -> int:
        for i, soil in enumerate(self.soil):
            if soil.name == key:
                return i
        raise KeyError(f"Soil with {key} not present.")

    def __getitem__(self, key) -> Soil:
        if isinstance(key, int):
            return self.soil[key]
        elif isinstance(key, str):
            for soil in self.soil:
                if soil.name == key:
                    return soil
            raise KeyError(key)


class Layer(DSeriesTreeStructure):
    name: str = ""  # will be generated by template
    material: int
    top_level: float  # [m]
    excess_pore_pressure_top: float = 0.0  # [kN/m3]
    excess_pore_pressure_bottom: float = 0.0  # [kN/m3]
    ocr_value: float = 1.0  # [-]
    reduction_core_resistance: int = 0  # [%]


class ReductionCoreResistanceEnum(IntEnum):
    SAFE = 0
    BEGEMANN = 1
    MANUAL = 2


class Profile(DSeriesTreeStructure):
    name: str
    matching_cpt: int
    x_coordinate: float
    y_coordinate: float
    phreatic_level: float
    pile_tip_level: float
    overconsolidation_ratio: float = 1.0
    top_of_positive_skin_friction: float
    bottom_of_negative_skin_friction: float
    expected_ground_level_settlement: float = 0.0
    placement_depth_of_foundation: float = 0.0
    concentration_value_frohlich: int = 3
    top_tension_zone: float = 0.0

    # Excavation part
    reduction_of_core_resistance: ReductionCoreResistanceEnum = (
        ReductionCoreResistanceEnum.SAFE
    )
    excavation_level: float
    excavation_width_infinite: Bool = Bool.TRUE
    excavation_length_infinite: Bool = Bool.TRUE
    distance_edge_pile_to_excavation_boundary: float = 0.0  # only valid for BEGEMANN

    layers: List[Layer] = []


class Profiles(DSeriesTreeStructureCollection):
    profiles: List[Profile] = []

    def add_profile_if_unique(self, profile: Profile) -> Profile:
        for added_profile in self.profiles:
            added_profile.excavation_level = profile.excavation_level
            if profile.name == added_profile.name:
                raise NameError(f"profile with name {profile.name} already exists.")
        self.profiles.append(profile)
        return profile


class InternalPile(BaseDataClass):
    # Only share method here, as shared properties
    # will not be picked up by parsers/pydantic
    def __eq__(self, other):
        """Overrides the default implementation"""
        if self.pile_name == other.pile_name:
            Warning(
                "Pile with name: "
                + self.pile_name
                + " already exists, pile is not added."
            )
            return True
        else:
            return False


class PositionBearingPile(InternalPile):
    index: PositiveInt = 1
    x_coordinate: confloat(ge=-10000000, le=100000000)
    y_coordinate: confloat(ge=-10000000, le=100000000)
    pile_head_level: confloat(ge=-1000, le=1000)
    surcharge: confloat(ge=0, le=100000)
    limit_state_str: confloat(ge=0, le=100000)
    limit_state_service: confloat(ge=0, le=100000)
    pile_name: constr(min_length=1, max_length=10)


class PositionTensionPile(InternalPile):
    index: PositiveInt = 1
    x_coordinate: confloat(ge=-10000000, le=100000000)
    y_coordinate: confloat(ge=-10000000, le=100000000)
    pile_head_level: confloat(ge=-1000, le=1000)
    use_alternating_loads: Bool
    max_force: confloat(ge=-100000, le=100000)
    min_force: confloat(ge=-100000, le=100000)
    limit_state_str: confloat(ge=0, le=100000)
    limit_state_service: confloat(ge=0, le=100000)
    pile_name: constr(min_length=1, max_length=10)


class PositionsBearingPiles(DSeriesNoParseSubStructure):
    positions: List[PositionBearingPile] = []


class PositionsTensionPiles(DSeriesNoParseSubStructure):
    positions: List[PositionTensionPile] = []


class CPTMeasureData(DFoundationsTableWrapper):
    data: List[Dict[str, float]]


class ExcavationType(IntEnum):
    BEFORE = 1
    AFTER = 2


class TimeOrderType(IntEnum):
    """Use this option to specify the execution time of CPTs relative to the pile installation.
    This information is needed to determine whether the problem qualifies for certain exceptions made in NEN 9997-1:2016.

    """

    CPT_EXCAVATION_INSTALL = 1
    INSTALL_CPT_EXCAVATION = 2
    EXCAVATION_CPT_INSTALL = 3
    EXCAVATION_INSTALL_CPT = 4
    INSTALL_EXCAVATION_CPT = 5
    CPT_INSTALL_EXCAVATION = 6
    CPT_BEFORE_AND_AFTER_INSTALL = 7


class InterpretationType(IntEnum):
    NEN_RULE = 0
    CUR = 1
    THREE_TYPE_RULE = 2
    QC_ONLY = 3
    UNKNOWN = 4


class CPT(DFoundationsEnumStructure):
    cptname: str
    project_name: str = UNKNOWN
    projectid: str = ""
    project_number: str = ""
    project_subnumber: str = ""
    location_name: str = UNKNOWN
    client_name: str = UNKNOWN
    companyid: str = UNKNOWN
    filedate: str = ""
    fileowner: str = UNKNOWN
    gef_version: str = UNKNOWN
    procedurecode: str = UNKNOWN
    objectid: str = "0"
    startdate: str = ""
    starttime: str = ""
    excavation_type: ExcavationType  # 1 : Before
    timeorder_type: TimeOrderType  # 1 : CPT - Excavation - Install
    cpt_type: int = 1  # Electronic
    usage_cone_value: int = 0  # Mechanical qc required
    xy_coordinate_system: str = UNKNOWN
    xworld: float = 987654321.00000
    xworld_accuracy: float = 987654321.00000
    yworld: float = 987654321.00000
    yworld_accuracy: float = 987654321.00000
    groundlevel: float
    groundlevel_accuracy: float = 987654321.00000
    groundlevel_was_measured: Bool = Bool.FALSE
    leveltext: str = ""
    pre_excavation: float
    waterlevel: float = 987654321.000
    xlocal: float = 0.0
    ylocal: float = 0.0
    local_x_crosssection: float = 0.000
    interpretation_model: InterpretationType
    interpretation_model_stressdependent: Bool
    depthrange: float
    graph_max_percentage: int = 10
    graph_width: float = 10.0
    graph_linewidth: int = 1
    graph_borderwidth: int = 1
    graph_bordercolor: int = 0
    graph_frictioncolor: int = 16711680
    graph_qccolor: int = 255
    graph_plane_color: int = 0
    graph_fit_for_size: Bool = Bool.TRUE
    graph_fit_symbol_for_size: Bool = Bool.TRUE
    graph_symbol_size: float = 1.0
    void_value_depth: float = 987654321.000
    void_value_cone_resistance: float = 987654321.000
    void_value_pore_water_pressure: float = 987654321.000
    void_value_sleeve_friction: float = 987654321.000
    void_value_friction_number: float = 987654321.000
    void_value_equivalent_electronic_qc: float = 987000000.000000
    measured_data: CPTMeasureData


class CPTList(DFoundationsCPTCollectionWrapper):
    cpt_collection: List[CPT] = []

    def add_cpt(self, cpt: CPT):
        """Add CPT and return id.
        Forces timeordertype to be the same everywhere.
        """
        for existing_cpt in self.cpt_collection:
            existing_cpt.timeorder_type = cpt.timeorder_type
        self.cpt_collection.append(cpt)
        return len(self.cpt_collection) - 1

    def __getitem__(self, key) -> CPT:
        if isinstance(key, int):
            return self.cpt_collection[key]
        elif isinstance(key, str):
            for cpt in self.cpt_collection:
                if cpt.cptname == key:
                    return cpt
            raise KeyError(key)


class CalculationOptions(DSeriesNoParseSubStructure):

    # Rigidity of superstructure
    is_rigid: Bool = Bool.TRUE

    # Transformation
    max_allowed_settlement_lim_state_str: confloat(ge=0, le=100000) = 0
    max_allowed_rel_rotation_lim_state_str: confloat(ge=0.0001, le=10000) = 0.0001
    max_allowed_settlement_lim_state_serv: confloat(ge=0, le=100000) = 0
    max_allowed_rel_rotation_lim_state_serv: confloat(ge=0.0001, le=10000) = 0.0001

    # Overrule parameters Bearing Piles
    is_xi3_overruled: Bool = Bool.FALSE
    factor_xi3: Optional[confloat(ge=0.01, le=10)] = 2
    is_xi4_overruled: Bool = Bool.FALSE
    factor_xi4: Optional[confloat(ge=0.01, le=10)] = 2
    is_gamma_b_overruled: Bool = Bool.FALSE
    factor_gamma_b: Optional[confloat(ge=1, le=100)] = 2
    is_gamma_s_overruled: Bool = Bool.FALSE
    factor_gamma_s: Optional[confloat(ge=1, le=100)] = 2
    is_gamma_fnk_overruled: Bool = Bool.FALSE
    factor_gamma_fnk: Optional[confloat(ge=-100, le=100)] = 2
    is_area_overruled: Bool = Bool.FALSE
    area: Optional[confloat(ge=0, le=100000)] = 1000
    is_ea_gem_overruled: Bool = Bool.FALSE
    ea_gem: Optional[confloat(ge=1)] = 100000

    # Model options combined
    is_suppress_qc_reduction: Bool = Bool.FALSE
    is_overrule_excavation: Bool = Bool.FALSE
    use_pile_group: Bool = Bool.TRUE
    is_write_intermediate_results: Bool = Bool.FALSE
    use_interaction_model: Bool = Bool.FALSE
    use_almere_rules: Bool = Bool.FALSE
    use_extra_almere_rules: Bool = Bool.FALSE

    # Overrule parameters Shallow foundations
    is_gamma_g_str_overruled: Bool = Bool.FALSE
    factor_gamma_g_str: Optional[confloat(ge=0.01, le=100)] = 1.0
    is_gamma_coh_overruled: Bool = Bool.FALSE
    factor_gamma_coh: Optional[confloat(ge=0.01, le=100)] = 1.0
    is_gamma_phi_overruled: Bool = Bool.FALSE
    factor_gamma_phi: Optional[confloat(ge=0.01, le=100)] = 1.0
    is_gamma_fundr_overruled: Bool = Bool.FALSE
    factor_gamma_fundr: Optional[confloat(ge=0.01, le=100)] = 1.0
    is_gamma_g_sls_overruled: Bool = Bool.FALSE
    factor_gamma_g_sls: Optional[confloat(ge=0.01, le=100)] = 1.0
    is_gamma_cc_overruled: Bool = Bool.FALSE
    factor_gamma_cc: Optional[confloat(ge=0.01, le=100)] = 1.0
    is_gamma_ca_overruled: Bool = Bool.FALSE
    factor_gamma_ca: Optional[confloat(ge=0.01, le=100)] = 1.0
    is_keep_length_constant: Bool = Bool.FALSE
    use_5_percent_limit: Bool = Bool.FALSE
    load_factor_between_limit_1_and_2: confloat(ge=0.5, le=1) = 0.833

    # Overrule parameters Tension Piles
    unit_weight_water: confloat(ge=0.01, le=20) = 9.81
    use_compaction: Bool = Bool.FALSE
    is_gamma_var_overruled: Bool = Bool.FALSE
    factor_gamma_var: Optional[confloat(ge=0.01, le=100)] = 1.0
    is_gamma_st_overruled: Bool = Bool.FALSE
    factor_gamma_st: Optional[confloat(ge=0.01, le=100)] = 1.0
    is_gamma_gamma_overruled: Bool = Bool.FALSE
    factor_gamma_gamma: Optional[confloat(ge=0.01, le=100)] = 1.0
    surcharge: confloat(ge=0, le=1e7) = 0
    use_piezometric_levels: Bool = Bool.TRUE

    # Model options Bearing Piles BE
    is_xi3_bpb_overruled: Bool = Bool.FALSE
    factor_xi3_bpb: confloat(ge=0.01, le=10) = 1.0
    is_xi4_bpb_overruled: Bool = Bool.FALSE
    factor_xi4_bpb: confloat(ge=0.01, le=10) = 1.0
    is_gamma_rd_overruled: Bool = Bool.FALSE
    factor_gamma_rd: confloat(ge=1, le=100) = 1.0
    is_gamma_b_bpb_overruled: Bool = Bool.FALSE
    factor_gamma_b_bpb: confloat(ge=1, le=100) = 1.0
    is_gamma_s_bpb_overruled: Bool = Bool.FALSE
    factor_gamma_s_bpb: confloat(ge=1, le=100) = 1.0
    suppress_all_qc_reduction_bp: Bool = Bool.FALSE
    use_quality_assurance: Bool = Bool.FALSE
    area_covered_per_cpt: confloat(ge=0) = 1000
    is_beta_bpb_overruled: Bool = Bool.FALSE
    factor_beat_bpb: confloat(ge=0.0, le=10) = 0

    def __init__(self, *args, **kwargs):
        """If defaults are overriden, update
        the related toggle fields as well.
        """
        toggles = {}
        for field, value in kwargs.items():
            if value is None:
                continue  # Nones will be passed by default settings
            toggle_field = self.find_toggle(field)
            if toggle_field in self.__fields__:
                toggles[toggle_field] = Bool.TRUE
        kwargs.update(toggles)
        super().__init__(*args, **kwargs)

    @staticmethod
    def find_toggle(field):
        """Transform given field like `factor_gamma_s_bpb`
        into `is_gamma_s_bpb_overruled`.
        """
        return "is_" + field.replace("factor_", "") + "_overruled"


class ModelTypeEnum(IntEnum):
    BEARING_PILES = 0
    TENSION_PILES = 1
    SHALLOW_FOUNDATIONS = 2
    BEARING_PILES_BELGIAN = 3


class ModelType(DFoundationsInlineProperties):
    model: ModelTypeEnum = ModelTypeEnum.BEARING_PILES

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # We only support Bearing & Tension Piles (NL)
        if self.model >= 2:
            logger.error(f"Model Type {self.model} is unsupported!")


class MainCalculationType(IntEnum):
    PRELIMINARY_DESIGN = 0
    VERIFICATION_DESIGN = 1


class SubCalculationType(IntEnum):
    VERIFICATION_DESIGN = 0
    VERIFICATION_COMPLETE = 1
    INDICATION_BEARING_CAPACITY = 2
    BEARING_CAPACITY_AT_FIXED_PILETIP_LEVELS = 3
    PILETIP_LEVELS_AND_NET_BEARING_CAPACITY = 4
    UNSUPPORTED = 7


class CalculationType(DFoundationsInlineProperties):
    main_calculationtype: MainCalculationType = MainCalculationType.PRELIMINARY_DESIGN
    sub_calculationtype: SubCalculationType = (
        SubCalculationType.INDICATION_BEARING_CAPACITY
    )

    def __init__(self, *args, **kwargs):
        # Set maintype automatically based on subtype
        super().__init__(*args, **kwargs)
        if self.sub_calculationtype.value >= 2:
            self.main_calculationtype = MainCalculationType.PRELIMINARY_DESIGN
        else:
            self.main_calculationtype = MainCalculationType.VERIFICATION_DESIGN


class PreliminaryDesign(DSeriesNoParseSubStructure):
    trajectory_begin: float = -10.00
    trajectory_end: float = -25.00
    trajectory_interval: float = 0.50

    profiles: List[int] = []  # ids of Profiles
    # Can be single pile type only in case of verification
    pile_types: List[int] = []  # ids of Piles

    # Only valid for Verification Design
    cpt_test_level: Optional[float] = 0.0  # [m]

    # Only valid for Prelimary Design
    net_bearing_capacity: Optional[int] = 0  # [kN]


class Version(DSerieVersion):
    soil: int = 1005
    d__foundations: int = 1015


class VersionExternal(DSeriesInlineMappedProperties):
    dgsfoundationcalc____dll: str = "19.1.2.26122"


class DFoundationsInputStructure(DSeriesStructure):
    """Representation of complete .foi file."""

    version: Version = Version()
    version_externals: VersionExternal = VersionExternal()
    model: ModelType = ModelType()
    soil_collection: SoilCollection = SoilCollection()
    run_identification: str = 6 * "\n"
    cpt_list: CPTList = CPTList()
    profiles: Profiles = Profiles()
    slopes: str = cleandoc(
        """
            0 = number of items
        """
    )
    types___bearing_piles: Union[List[TypesBearingPiles], str] = cleandoc(
        """
        -1 : pile type shown in main graph
            0 = number of items
        """
    )
    types___bearing_piles_belgian: str = cleandoc(
        """
        -1 : pile type shown in main graph
            0 = number of items
        """
    )
    types___tension_piles_cur: Union[List[TypesTensionPiles], str] = cleandoc(
        """
        -1 : pile type shown in main graph
            0 = number of items
        """
    )
    types___shallow_foundations: str = cleandoc(
        """
            0 = number of items
        """
    )
    loads: str = cleandoc(
        """
            0 = number of items
        """
    )
    positions___bearing_piles: Union[PositionsBearingPiles, str] = PositionsBearingPiles()
    positions___bearing_piles_belgian: str = cleandoc(
        """
        [TABLE]
        [COLUMN INDICATION]
        index
        X
        Y
        PileHeadLevel
        Surcharge
        LimitStateStrGeo
        LimitStateService
        PileName
        [END OF COLUMN INDICATION]
        [DATA]
        [END OF DATA]
        [END OF TABLE]
        """
    )
    positions___tension_piles_cur: Union[
        PositionsTensionPiles, str
    ] = PositionsTensionPiles()
    positions___shallow_foundations: str = cleandoc(
        """
        [TABLE]
        [COLUMN INDICATION]
        index
        X
        Y
        Angle
        FoundationType
        Load
        Profile
        Slope
        PileName
        [END OF COLUMN INDICATION]
        [DATA]
        [END OF DATA]
        [END OF TABLE]
        """
    )
    calculation_options: Union[CalculationOptions, str] = CalculationOptions()
    calculationtype: CalculationType = CalculationType()
    preliminary_design: Union[PreliminaryDesign, str] = PreliminaryDesign()
    de_beer: str = cleandoc(
        """
          0.0357 : Cone diameter [m]
            1.00 : Installation factor [-]
           1.000 : Scale factor [-]
           -4.00 : Trajectory begin [m]
          -20.00 : Trajectory end [m]
            1.00 : Trajectory interval [m]
        0 : Number of Profiles selected for calculation
            -1 : No pile type selected
        0 : Lambda overruled = FALSE
        """
    )
    location_map: str = cleandoc(
        """
         0.0000
                0.0000
                0.0000
                0.0000
        """
    )

    # Custom validator
    _validate_run_identification = make_newline_validator(
        "run_identification", req_newlines=REQ_RUN_LINES
    )


class DFoundationsNenPileResultsTable(DFoundationsTableWrapper):
    data: List[Dict[str, Union[int, float, str]]] = []


class DFoundationsCalculationParametersBearingPilesEC7(DSeriesInlineProperties):
    ksi3used: float
    ksi4used: float
    gammabused: float
    gammasused: float
    isksi3used: bool


# region VerificationResults


class DFoundationsNenPileResults(DFoundationsInlineProperties):
    cpts: DFoundationsNenPileResultsTable
    pile_point_shape_factor: float
    factor_influence_crosssection_of_pile_point_s: float
    max_load_on_foundation: float
    max_bearing_capacity_foundation: float
    critical_pile_nr_for_settlement_in_gt1b: int
    critical_cpt_nr_for_settlement_in_gt1b: int
    critical_pile_nr_for_settlement_in_gt2: int
    critical_cpt_nr_for_settlement_in_gt2: int
    first_pile_nr_for_critical_rotation_in_gt1b: int
    second_pile_nr_for_critical_rotation_in_gt1b: int
    first_pile_nr_for_critical_rotation_in_gt2: int
    second_pile_nr_for_critical_rotation_ingt2: int
    fs_tot_d_1b: float
    fr_max_punt_d_1b: float
    fr_max_schacht_d_1b: float
    fr_punt_d_1b: float
    spunt_d_1b: float
    sel_d1b: float
    s2_d1b: float
    sneg1b: float
    fs_tot_d_2: float
    fr_max_punt_d2: float
    fr_max_schacht_d_2: float
    fr_punt_d_2: float
    spunt_d_2: float
    sel_d2: float
    s2_d2: float
    sneg2: float
    max_shaft_and_point: DFoundationsNenPileResultsTable
    sigma_max_schacht_1b: float
    sigma_max_schacht_2: float
    min_value: float
    max_value: float
    nen_average_pile_factors: DFoundationsNenPileResultsTable
    calculation_parameters_bearing_piles_ec_7: DFoundationsCalculationParametersBearingPilesEC7

    @classmethod
    def header_lines(cls) -> int:
        """Tells to the parent class that this structure text comes
        with three lines of header text that can be skipped.

        Returns:
            int: Header line size.
        """
        return 3


class DFoundationsGlobalNenResults(DFoundationsInlineProperties):
    wd1b: float
    betad1b: float
    w2d: float
    betad2: float


class DFoundationsVerificationResults(DSeriesStructure):
    global_nen_results: Optional[DFoundationsGlobalNenResults]
    demands_nen__en: Optional[str]
    nen_pile_results: Optional[DFoundationsNenPileResults]

    verification_results_tp_load__settlement_curve_1b: Optional[str]
    verification_results_tp_1a: Optional[str]
    verification_results_tp_1b2: Optional[str]
    verification_results_tp_load__settlement_curve_2: Optional[str]


# endregion


# region Output


class DFoundationsCalculationWarnings(DSeriesTreeStructure):
    is_warning_f1_given: Bool
    f1_greater_than_1_found: Bool
    is_warning_nen_depth_given: Bool
    is_warning_sf_fund_width_given: Bool
    is_warning_sf_fund_length_given: Bool
    is_warning_sf_cud_given: Bool
    is_warning_sf_delta_phi_given: Bool
    is_warning_nen_spacing_given: int = int()
    is_warning_sf_slope_not_relevant_given: Bool
    is_warning_nen_sf_placement_depth_too_deep: Bool
    is_warning_nen_sf_placement_depth_too_shallow: Bool


class DFoundationsDumpfileOutputStructure(DSeriesStructure):
    results_at_cpt_test_level: Optional[str]
    verification_results: Optional[DFoundationsVerificationResults]

    calculation_parameters_tension_piles: Optional[str]
    verification_results_tp: Optional[DFoundationsVerificationResults]

    footnote_warnings: Optional[str]
    preliminary_design_results: Optional[str]
    verification_results_sf: Optional[str]
    verification_results_tp_1b2: Optional[str]
    verification_design_results: Optional[str]


class DFoundationsStructure(DSeriesStructure):
    input_data: DFoundationsInputStructure = DFoundationsInputStructure()
    dumpfile_output: Optional[DFoundationsDumpfileOutputStructure]
    calculation_warnings: Optional[DFoundationsCalculationWarnings]


class DFoundationsDumpStructure(DSeriesStructure):
    dumpfile: DFoundationsStructure = DFoundationsStructure()


# endregion



from abc import ABCMeta, abstractmethod
from typing import Optional

from pydantic import conlist, constr, validator

from geolib.geometry import Point
from geolib.models import BaseDataClass

from .internal import HorizontalLineLoad as HorizontalLineLoadInternal
from .internal import Moment as MomentInternal
from .internal import NormalForce as NormalForceInternal
from .internal import SurchargeLoad as InternalSurchargeLoad
from .internal import SurchargePoint as InternalSurchargePoint
from .internal import UniformLoad as InternalUniformLoad
from .settings import (
    DistributionType,
    LoadTypeFavourableUnfavourable,
    LoadTypeFavourableUnfavourableMoment,
    LoadTypePermanentVariable,
)


class VerificationLoadSettingsHorizontalLineLoad(BaseDataClass):
    """
    These options are available only if the Verification (EC7/CUR) option is selected in the Model window for the D-SheetPiling model.
    For GEOLIB the "verification" parameter should be set to True in the function DSheetPilingModel.set_model

    Args:
        duration_type: Select the duration of load application, Permanent or Variable.
        load_type: Select the type of load, Favorable, Unfavorable or leave it D-Sheet Piling determined.
    """

    load_type: LoadTypeFavourableUnfavourable = (
        LoadTypeFavourableUnfavourable.DSHEETPILING_DETERMINED
    )
    duration_type: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT


class VerificationLoadSettingsLoads(BaseDataClass):
    """
    These options are available only if the Verification (EC7/CUR) option is selected in the Model window for the D-SheetPiling model.
    For GEOLIB the "verification" parameter should be set to True in the function DSheetPilingModel.set_model

    Args:
        duration_type: Select the duration of load application, Permanent or Variable.
        load_type: Select the type of load, Favorable, Unfavorable or leave it D-Sheet Piling determined.
    """

    load_type: LoadTypeFavourableUnfavourable = (
        LoadTypeFavourableUnfavourable.DSHEETPILING_DETERMINED
    )
    duration_type: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT


class VerificationLoadSettings(BaseDataClass):
    """
    Load class for moment loads
    These options are available only if the Verification (EC7/CUR) option is selected in the Model window for the D-SheetPiling model.
    For GEOLIB the "verification" parameter should be set to True in the function DSheetPilingModel.set_model

    Args:
        duration_type: Select the duration of load application, Permanent or Variable.
        load_type: Select the type of load, Favorable or Unfavorable.
    """

    load_type: LoadTypeFavourableUnfavourableMoment = (
        LoadTypeFavourableUnfavourableMoment.FAVOURABLE
    )
    duration_type: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT


class VerificationLoadSettingsMomentNormalForce(BaseDataClass):
    """
    Load class for moment loads
    These options are available only if the Verification (EC7/CUR) option is selected in the Model window for the D-SheetPiling model.
    For GEOLIB the "verification" parameter should be set to True in the function DSheetPilingModel.set_model

    Args:
        duration_type: Select the duration of load application, Permanent or Variable.
        load_type: Select the type of load, Favorable or Unfavorable.
    """

    load_type: LoadTypeFavourableUnfavourableMoment = (
        LoadTypeFavourableUnfavourableMoment.FAVOURABLE
    )
    duration_type: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT


class UniformLoad(BaseDataClass):
    """Uniform Load. This option is not available for SignlePileModelType.

    Args:
        name: Name of the SurchargeLoad.
        left_load: Value of the load on the left side.
        right_load: Value of the load on the right side.
        verification_load_settings: VerificationLoadSettings.
        standard_deviation_left: Standard deviation of the distribution type of the left side.
        standard_deviation_right: Standard deviation of the distribution type of the right side.
        distribution_type_left: Distribution type of the left side.
        distribution_type_right: Distribution type of the right side.
    """

    name: constr(min_length=1, max_length=50)
    left_load: float
    right_load: float
    verification_load_settings: VerificationLoadSettingsLoads = (
        VerificationLoadSettingsLoads()
    )
    standard_deviation_left: float = 0.0
    standard_deviation_right: float = 0.0
    distribution_type_left: DistributionType = DistributionType.NORMAL
    distribution_type_right: DistributionType = DistributionType.NORMAL

    def to_internal(self) -> InternalUniformLoad:
        uniformload = InternalUniformLoad(
            name=self.name,
            uniformloadleft=self.left_load,
            uniformloadright=self.right_load,
            uniformloadstandarddeviationleft=self.standard_deviation_left,
            uniformloadstandarddeviationright=self.standard_deviation_right,
        )
        uniformload.uniformloadpermanent = self.verification_load_settings.duration_type
        uniformload.uniformloadfavourable = self.verification_load_settings.load_type
        if self.distribution_type_left:
            uniformload.uniformloaddistleft = self.distribution_type_left
        if self.distribution_type_right:
            uniformload.uniformloaddistright = self.distribution_type_right
        return uniformload


class Moment(BaseDataClass):
    """Moment Load."""

    name: constr(min_length=1, max_length=50)
    level: float
    load: float
    verification_load_settings: VerificationLoadSettingsMomentNormalForce = (
        VerificationLoadSettingsMomentNormalForce()
    )

    def to_internal(self) -> MomentInternal:
        moment = MomentInternal(
            **self.dict(exclude_none=True, exclude={"verification_load_settings"})
        )
        moment.load_type = self.verification_load_settings.load_type
        moment.duration_type = self.verification_load_settings.duration_type
        return moment


class SurchargeLoad(BaseDataClass):
    """Surcharge Load.

    Args:
        name: Name of the SurchargeLoad.
        points: SurchargeLoad points. Points (X, Z) must be defined from the sheet outwards regarding x-coordinate.
        verification_load_settings: VerificationLoadSettings.
        standard_deviation: Standard deviation of the distribution type.
        distribution_type: Distribution type.
    """

    name: constr(min_length=1, max_length=50)
    points: conlist(Point, min_items=1)
    verification_load_settings: VerificationLoadSettingsLoads = (
        VerificationLoadSettingsLoads()
    )
    standard_deviation: float = 0.0
    distribution_type: DistributionType = DistributionType.NORMAL

    @validator("points")
    def points_must_be_increasing_and_greater_or_equal_to_zero(cls, v):
        x_coords = [p.x for p in v]
        if min(x_coords) < 0:
            raise ValueError(
                f"All x-coordinates must be greater than or equal to 0, found {min(x_coords)}"
            )
        if x_coords != sorted(x_coords):
            raise ValueError("x-coordinates must be strictly increasing")
        return v

    def to_internal(self) -> InternalSurchargeLoad:
        surchargeload = InternalSurchargeLoad(
            name=self.name,
            points=[
                InternalSurchargePoint(
                    surchargeloaddistance=point.x, surchargeloadvalue=point.z
                )
                for point in self.points
            ],
        )
        surchargeload.surchargeloadpermanent = (
            self.verification_load_settings.duration_type
        )
        surchargeload.surchargeloadfavourable = self.verification_load_settings.load_type
        if self.standard_deviation:
            surchargeload.surchargeloadstandarddeviation = self.standard_deviation
        if self.distribution_type:
            surchargeload.surchargeloaddistribution = self.distribution_type
        return surchargeload


class HorizontalLineLoad(BaseDataClass):
    """Horizontal Line Load."""

    name: constr(min_length=1, max_length=50)
    level: float
    load: float
    verification_load_settings: VerificationLoadSettingsLoads = (
        VerificationLoadSettingsLoads()
    )

    def to_internal(self) -> HorizontalLineLoadInternal:
        horizontallineload = HorizontalLineLoadInternal(
            **self.dict(exclude_none=True, exclude={"verification_load_settings"})
        )
        horizontallineload.load_type = self.verification_load_settings.load_type
        horizontallineload.duration_type = self.verification_load_settings.duration_type
        return horizontallineload


class NormalForce(BaseDataClass):
    """Normal Force Load."""

    name: constr(min_length=1, max_length=50)
    force_at_sheet_pile_top: float
    force_at_surface_level_left_side: float
    force_at_surface_level_right_side: float
    force_at_sheet_pile_toe: float
    verification_load_settings: VerificationLoadSettingsMomentNormalForce = (
        VerificationLoadSettingsMomentNormalForce()
    )

    def to_internal(self) -> NormalForceInternal:
        normalforce = NormalForceInternal(
            **self.dict(exclude_none=True, exclude={"verification_load_settings"})
        )
        normalforce.load_type = self.verification_load_settings.load_type
        normalforce.duration_type = self.verification_load_settings.duration_type
        return normalforce


class SoilDisplacement(BaseDataClass):
    """Non Uniform Load."""


class Earthquake(BaseDataClass):
    """Non Uniform Load."""

    force: float  # g



# -*- coding: utf-8 -*-

"""
This module contains the primary objects that power GEOLib.
"""
import abc
import logging
import os
from abc import abstractmethod, abstractproperty
from pathlib import Path, PosixPath, WindowsPath
from subprocess import Popen, run
from types import CoroutineType
from typing import List, Optional, Type, Union

import requests
from pydantic import DirectoryPath, FilePath, HttpUrl, conlist
from pydantic.error_wrappers import ValidationError
from requests.auth import HTTPBasicAuth

from geolib.errors import CalculationError
from geolib.models import BaseDataClass

from .base_model_structure import BaseModelStructure
from .meta import MetaData
from .parsers import BaseParserProvider

logger = logging.getLogger(__name__)
meta = MetaData()


class BaseModel(BaseDataClass, abc.ABC):
    filename: Optional[Path]
    datastructure: Optional[BaseModelStructure]
    meta: MetaData = MetaData()

    def execute(self, timeout_in_seconds: int = meta.timeout) -> "BaseModel":
        """Execute a Model and wait for `timeout` seconds.

        The model is modified in place if the calculation and parsing
        is successful.
        """
        if self.filename is None:
            raise ValueError("Set filename or serialize first!")
        if not self.filename.exists():
            logger.warning("Serializing before executing.")
            self.serialize(self.filename)

        executable = self.meta.console_folder / self.console_path
        if not executable.exists():
            logger.error(
                f"Please make sure the `geolib.env` file points to the console folder. GEOLib now can't find it at `{executable}`"
            )
            raise CalculationError(-1, f"Console executable not found at {executable}.")

        process = run(
            [str(executable)]
            + self.console_flags
            + [str(self.filename.resolve())]
            + self.console_flags_post,
            timeout=timeout_in_seconds,
            cwd=str(self.filename.resolve().parent),
        )
        logger.debug(f"Executed with {process.args}")

        # Successfull run
        output_filename = output_filename_from_input(self)
        logger.info(
            f"Checking for {output_filename}, while process exited with {process.returncode}"
        )
        if output_filename.exists():
            try:
                self.parse(output_filename)
                return self  # TODO Figure out whether we should instantiate a new model (parse is a classmethod)
            except ValidationError:
                logger.warning(
                    f"Ouput file generated but parsing of {output_filename} failed."
                )
                error = self.get_error_context()
                raise CalculationError(process.returncode, error)

        # Unsuccessful run
        else:
            error = self.get_error_context()
            raise CalculationError(process.returncode, error)

    def execute_remote(self, endpoint: HttpUrl) -> "BaseModel":
        """Execute a Model on a remote endpoint.

        A new model instance is returned.
        """
        response = requests.post(
            requests.compat.urljoin(
                endpoint, f"calculate/{self.__class__.__name__.lower()}"
            ),
            data=self.json(),
            auth=HTTPBasicAuth(self.meta.gl_username, self.meta.gl_password),
        )
        if response.status_code == 200:
            data = response.json()
            # remove possibly invalid external metadata
            data.get("meta", {}).pop("console_folder", None)
            return self.__class__(**data)
        else:
            raise CalculationError(response.status_code, response.text)

    def get_error_context(self) -> str:
        err_fn = output_filename_from_input(self, extension=".err")
        batch_fn = self.filename.parent / "Batchlog.txt"
        error = f"{self.filename.name}\n"
        if err_fn.exists():
            with open(err_fn) as f:
                error += f"### {err_fn} ###\n"
                error += f.read()
        elif batch_fn.exists():
            with open(batch_fn) as f:
                error += f"### {batch_fn} ###\n"
                error += f.read()
        else:
            error = f"Couldn't determine source of error for {self.filename.name}."
        return error

    @abstractmethod
    def serialize(
        self, filename: Union[FilePath, DirectoryPath, None]
    ) -> Union[FilePath, DirectoryPath, None]:
        """Serialize model to input file."""

    @property
    def console_path(self) -> Path:
        raise NotImplementedError("Implement in concrete classes.")

    @property
    def console_flags(self) -> List[str]:
        return []

    @property
    def console_flags_post(self) -> List[str]:
        return []

    @abstractproperty
    def parser_provider_type(self) -> Type[BaseParserProvider]:
        """Returns the parser provider type of the current concrete class.

        Raises:
            NotImplementedError: If not implemented in the concrete class.

        Returns:
            Type[BaseParserProvider] -- Concrete parser provider.
        """
        raise NotImplementedError("Implement in concrete classes.")

    def parse(self, filename: FilePath) -> BaseModelStructure:
        """Parse input or outputfile to Model, depending on extension."""
        self.filename = filename
        # self.datastructure = None
        self.datastructure = self.parser_provider_type().parse(filename)
        return self.datastructure

    @property
    def is_valid(self) -> bool:
        """Checks validity and integrity of structure."""
        if self.datastructure is not None:
            return self.datastructure.is_valid
        else:
            logger.warning("No datastructured parsed yet!")
            return False

    def set_metadata(self, meta: MetaData):
        """Set custom metadata for input file."""
        self.metadata = meta

    @property
    def input(self):
        """Access internal dict-like datastructure of the input."""
        return self.datastructure

    @property
    def output(self):
        """Access internal dict-like datastructure of the output.

        Requires a successful execute.
        """
        return self.datastructure.results


class BaseModelList(BaseDataClass):
    """Hold multiple models that can be executed in parallel.

    Note that all models need to have a unique filename
    otherwise they will overwrite eachother. This also helps with
    identifying them later."""

    models: List[BaseModel]
    meta: MetaData = MetaData()
    errors: List[str] = []

    def execute(
        self,
        calculation_folder: DirectoryPath,
        timeout_in_seconds: int = meta.timeout,
        nprocesses: Optional[int] = os.cpu_count(),
    ) -> "BaseModelList":
        """Execute all models in this class in parallel.

        We split the list to separate folders and call a batch processes on each folder.
        Note that the order of models will change.
        """

        # manual check as remote execution could result in zero models
        if len(self.models) == 0:
            raise ValueError("Can't execute with zero models.")

        lead_model = self.models[0]
        processes = []
        output_models = []
        errors = []

        # Divide the models over n processes and make sure to copy them to prevent aliasing
        split_models = [self.models[i::nprocesses] for i in range(nprocesses)]
        for i, models in enumerate(split_models):
            if len(models) == 0:
                continue
            unique_folder = calculation_folder / str(i)
            unique_folder.mkdir(parents=True, exist_ok=True)

            for model in models:
                fn = unique_folder / model.filename.name
                model.serialize(fn.resolve())

            executable = self.meta.console_folder / lead_model.console_path
            if not executable.exists():
                logger.error(
                    f"Please make sure the `geolib.env` file points to the console folder. GEOLib now can't find it at `{executable}`"
                )
                raise CalculationError(
                    -1, f"Console executable not found at {executable}."
                )

            process = Popen(
                [str(executable)] + lead_model.console_flags + [str(i)],
                cwd=str(calculation_folder.resolve()),
            )
            processes.append(process)

        # Wait for all processes to be done
        for process in processes:
            logger.debug(f"Executed with {process.args}")
            process.wait(timeout=timeout_in_seconds)

        # Iterate over the models
        for i, models in enumerate(split_models):
            for model in models:
                model = model.copy(deep=True)  # prevent aliasing
                output_filename = output_filename_from_input(model)
                if output_filename.exists():
                    try:
                        model.parse(output_filename)
                        output_models.append(model)

                    except ValidationError:
                        logger.warning(
                            f"Ouput file generated but parsing of {output_filename.name} failed."
                        )
                        error = model.get_error_context()
                        errors.append(error)
                else:
                    logger.warning(
                        f"Model @ {output_filename.name} failed. Please check the .err file and batchlog.txt in its folder."
                    )
                    error = model.get_error_context()
                    errors.append(error)

        return self.__class__(models=output_models, errors=errors)

    def execute_remote(self, endpoint: HttpUrl) -> "BaseModelList":
        """Execute all models in this class in parallel on a remote endpoint.

        Note that the order of models will change.
        """
        lead_model = self.models[0]

        response = requests.post(
            requests.compat.urljoin(
                endpoint, f"calculate/{lead_model.__class__.__name__.lower()}s"
            ),
            data="[" + ",".join((model.json() for model in self.models)) + "]",
            auth=HTTPBasicAuth(lead_model.meta.gl_username, lead_model.meta.gl_password),
        )
        if response.status_code == 200:
            models = response.json()["models"]
            errors = response.json()["errors"]
            stripped_models = []
            for model in models:
                # remove possibly invalid external metadata
                model.get("meta", {}).pop("console_folder", None)
                stripped_models.append(lead_model.__class__(**model))
            return self.__class__(models=stripped_models, errors=errors)
        else:
            raise CalculationError(response.status_code, response.text)


def output_filename_from_input(model: BaseModel, extension: str = None) -> Path:
    if not extension:
        extension = model.parser_provider_type().output_parsers[-1].suffix_list[0]
    return model.filename.with_suffix(extension)



from enum import IntEnum


class DrainType(IntEnum):
    STRIP = 0
    COLUMN = 1
    SANDWALL = 2


class DrainGridType(IntEnum):
    TRIANGULAR = 0
    RECTANGULAR = 1
    UNDERDETERMINED = 2


class DrainSchedule(IntEnum):
    OFF = 0
    SIMPLE_INPUT = 1
    DETAILED_INPUT = 2



from geolib.models import BaseDataClass
from geolib.models.dseries_parser import DSeriesInlineMappedProperties


class PartialFactorsEurocodeDa1Set1(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.00
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.35
    effectfactorvarunfav: float = 1.10
    materialfactorcohesion: float = 1.00
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.000
    materialfactorsubgradereaction: float = 1.00
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 10.00
    geometrymaxincretainingheight: float = 0.50
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.00
    geometrydeltaactivephreaticline: float = 0.00
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.00
    overallstabilityfactortgphi: float = 1.00
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 0.00
    verticalbalancegammamb: float = 1.25


class PartialFactorsEurocodeDa1Set2(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.30
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.25
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.250
    materialfactorsubgradereaction: float = 1.00
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 10.00
    geometrymaxincretainingheight: float = 0.50
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.00
    geometrydeltaactivephreaticline: float = 0.00
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.00
    overallstabilityfactortgphi: float = 1.00
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 0.00
    verticalbalancegammamb: float = 1.25


class PartialFactorsEurocodeDa2(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.00
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.35
    effectfactorvarunfav: float = 1.10
    materialfactorcohesion: float = 1.00
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.000
    materialfactorsubgradereaction: float = 1.00
    resistancefactor: float = 1.40
    geometryincretainingheight: float = 10.00
    geometrymaxincretainingheight: float = 0.50
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.00
    geometrydeltaactivephreaticline: float = 0.00
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.00
    overallstabilityfactortgphi: float = 1.00
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 0.00
    verticalbalancegammamb: float = 1.25


class PartialFactorsEurocodeDa3(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.30
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.25
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.250
    materialfactorsubgradereaction: float = 1.00
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 10.00
    geometrymaxincretainingheight: float = 0.50
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.00
    geometrydeltaactivephreaticline: float = 0.00
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.00
    overallstabilityfactortgphi: float = 1.00
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 0.00
    verticalbalancegammamb: float = 1.25


class PartialFactorsEc7Nl0(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.00
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.00
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.050
    materialfactorsubgradereaction: float = 1.30
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 10.00
    geometrymaxincretainingheight: float = 0.50
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.15
    geometrydeltaactivephreaticline: float = 0.05
    overallstabilityfactordrivingmoment: float = 0.90
    overallstabilityfactorcohesion: float = 1.30
    overallstabilityfactortgphi: float = 1.20
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 1.20
    verticalbalancegammamb: float = 1.20


class PartialFactorsEc7Nl1(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.00
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.15
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.150
    materialfactorsubgradereaction: float = 1.30
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 10.00
    geometrymaxincretainingheight: float = 0.50
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.20
    geometrydeltaactivephreaticline: float = 0.05
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.30
    overallstabilityfactortgphi: float = 1.20
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 1.20
    verticalbalancegammamb: float = 1.25


class PartialFactorsEc7Nl2(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.00
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.25
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.175
    materialfactorsubgradereaction: float = 1.30
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 10.00
    geometrymaxincretainingheight: float = 0.50
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.25
    geometrydeltaactivephreaticline: float = 0.05
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.45
    overallstabilityfactortgphi: float = 1.25
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 1.20
    verticalbalancegammamb: float = 1.25


class PartialFactorsEc7Nl3(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.25
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.40
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.200
    materialfactorsubgradereaction: float = 1.30
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 10.00
    geometrymaxincretainingheight: float = 0.50
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.25
    geometrydeltaactivephreaticline: float = 0.05
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.60
    overallstabilityfactortgphi: float = 1.30
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 1.35
    verticalbalancegammamb: float = 1.25


class PartialFactorsEc7BSet1(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.00
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.35
    effectfactorvarunfav: float = 1.10
    materialfactorcohesion: float = 1.00
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.000
    materialfactorsubgradereaction: float = 1.00
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 10.00
    geometrymaxincretainingheight: float = 0.50
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.00
    geometrydeltaactivephreaticline: float = 0.00
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.00
    overallstabilityfactortgphi: float = 1.00
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 0.00
    verticalbalancegammamb: float = 1.20


class PartialFactorsEc7BSet2(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.10
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.25
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.250
    materialfactorsubgradereaction: float = 1.00
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 10.00
    geometrymaxincretainingheight: float = 0.50
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.00
    geometrydeltaactivephreaticline: float = 0.00
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.25
    overallstabilityfactortgphi: float = 1.25
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 0.00
    verticalbalancegammamb: float = 1.20


class PartialFactorsCurI(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.00
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.00
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.050
    materialfactorsubgradereaction: float = 1.30
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 0.00
    geometrymaxincretainingheight: float = 0.00
    geometrydeltapassivesurfacelevel: float = 0.20
    geometrydeltapassivephreaticline: float = 0.15
    geometrydeltaactivephreaticline: float = 0.05
    overallstabilityfactordrivingmoment: float = 0.90
    overallstabilityfactorcohesion: float = 1.50
    overallstabilityfactortgphi: float = 1.20
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 0.00
    verticalbalancegammamb: float = 1.25


class PartialFactorsCurIi(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.00
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.00
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.150
    materialfactorsubgradereaction: float = 1.30
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 0.00
    geometrymaxincretainingheight: float = 0.00
    geometrydeltapassivesurfacelevel: float = 0.30
    geometrydeltapassivephreaticline: float = 0.20
    geometrydeltaactivephreaticline: float = 0.05
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.50
    overallstabilityfactortgphi: float = 1.20
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 0.00
    verticalbalancegammamb: float = 1.25


class PartialFactorsCurIii(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.25
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.00
    loadfactorvarunfavmultiply: float = 1.00
    loadfactorpermunfavcalc: float = 0.0000
    loadfactorvarunfavcalc: float = 0.0000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 1.00
    constructloadfactorpermunfavmultiply: float = 1.00
    constructloadfactorvarunfavmultiply: float = 1.00
    constructloadfactorpermunfavcalc: float = 1.0000
    constructloadfactorvarunfavcalc: float = 1.0000
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.10
    materialfactorundrainedshearstrength: float = 1.00
    materialfactortgphi: float = 1.200
    materialfactorsubgradereaction: float = 1.30
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 0.00
    geometrymaxincretainingheight: float = 0.00
    geometrydeltapassivesurfacelevel: float = 0.35
    geometrydeltapassivephreaticline: float = 0.25
    geometrydeltaactivephreaticline: float = 0.05
    overallstabilityfactordrivingmoment: float = 1.10
    overallstabilityfactorcohesion: float = 1.50
    overallstabilityfactortgphi: float = 1.20
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 0.00
    verticalbalancegammamb: float = 1.25


class PartialFactorsEc7SeVk1(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 0.83
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.00
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.10
    loadfactorvarunfavmultiply: float = 1.40
    loadfactorpermunfavcalc: float = 0.9130
    loadfactorvarunfavcalc: float = 1.1620
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 0.00
    constructloadfactorpermunfavmultiply: float = 1.35
    constructloadfactorvarunfavmultiply: float = 1.50
    constructloadfactorpermunfavcalc: float = 1.1205
    constructloadfactorvarunfavcalc: float = 1.2450
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.30
    materialfactorundrainedshearstrength: float = 1.50
    materialfactortgphi: float = 1.300
    materialfactorsubgradereaction: float = 1.00
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 0.00
    geometrymaxincretainingheight: float = 0.00
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.00
    geometrydeltaactivephreaticline: float = 0.00
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.30
    overallstabilityfactortgphi: float = 1.30
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 1.00
    verticalbalancegammamb: float = 1.30


class PartialFactorsEc7SeVk2(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 0.91
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.00
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.10
    loadfactorvarunfavmultiply: float = 1.40
    loadfactorpermunfavcalc: float = 1.0010
    loadfactorvarunfavcalc: float = 1.2740
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 0.00
    constructloadfactorpermunfavmultiply: float = 1.35
    constructloadfactorvarunfavmultiply: float = 1.50
    constructloadfactorpermunfavcalc: float = 1.2285
    constructloadfactorvarunfavcalc: float = 1.3650
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.30
    materialfactorundrainedshearstrength: float = 1.50
    materialfactortgphi: float = 1.300
    materialfactorsubgradereaction: float = 1.00
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 0.00
    geometrymaxincretainingheight: float = 0.00
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.00
    geometrydeltaactivephreaticline: float = 0.00
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.30
    overallstabilityfactortgphi: float = 1.30
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 1.00
    verticalbalancegammamb: float = 1.30


class PartialFactorsEc7SeVk3(DSeriesInlineMappedProperties):
    safetyfactorgammad: float = 1.00
    safetyfactoreta: float = 1.00
    loadfactorpermunfav: float = 1.00
    loadfactorpermfav: float = 1.00
    loadfactorvarunfav: float = 1.00
    loadfactorvarfav: float = 0.00
    loadfactorpermunfavmultiply: float = 1.10
    loadfactorvarunfavmultiply: float = 1.40
    loadfactorpermunfavcalc: float = 1.1000
    loadfactorvarunfavcalc: float = 1.4000
    constructloadfactorpermfav: float = 1.00
    constructloadfactorvarfav: float = 0.00
    constructloadfactorpermunfavmultiply: float = 1.35
    constructloadfactorvarunfavmultiply: float = 1.50
    constructloadfactorpermunfavcalc: float = 1.3500
    constructloadfactorvarunfavcalc: float = 1.5000
    effectfactor: float = 1.00
    effectfactorvarunfav: float = 1.00
    materialfactorcohesion: float = 1.30
    materialfactorundrainedshearstrength: float = 1.50
    materialfactortgphi: float = 1.300
    materialfactorsubgradereaction: float = 1.00
    resistancefactor: float = 1.00
    geometryincretainingheight: float = 0.00
    geometrymaxincretainingheight: float = 0.00
    geometrydeltapassivesurfacelevel: float = 0.00
    geometrydeltapassivephreaticline: float = 0.00
    geometrydeltaactivephreaticline: float = 0.00
    overallstabilityfactordrivingmoment: float = 1.00
    overallstabilityfactorcohesion: float = 1.30
    overallstabilityfactortgphi: float = 1.30
    overallstabilityfactorunitweight: float = 1.00
    factorrepvaluesmdpmax: float = 1.00
    verticalbalancegammamb: float = 1.30



import logging
from enum import Enum
from pathlib import Path
from subprocess import CompletedProcess, run
from typing import List, Optional, Type, Union

from pydantic import FilePath, confloat

from geolib.geometry import Point
from geolib.models import BaseDataClass, BaseModel, BaseModelStructure
from geolib.models.internal import Bool
from geolib.models.meta import CONSOLE_RUN_BATCH_FLAG
from geolib.soils import Soil

from .dfoundations_parserprovider import DFoundationsParserProvider
from .internal import CalculationOptions as InternalModelOptions
from .internal import CalculationType as InternalCalculationType
from .internal import (
    CPTList,
    DFoundationsDumpfileOutputStructure,
    DFoundationsDumpStructure,
    DFoundationsStructure,
    ModelTypeEnum,
    PositionsBearingPiles,
    PositionsTensionPiles,
    PreliminaryDesign,
    SoilCollection,
)
from .internal import SubCalculationType as CalculationType
from .internal_soil import Soil as InternalSoil
from .piles import (
    BearingPile,
    BearingPileLocation,
    Pile,
    PileLocation,
    TensionPile,
    TensionPileLocation,
)
from .profiles import Profile
from .serializer import DFoundationsInputSerializer

logger = logging.getLogger(__name__)


class ModelOptions(BaseDataClass):
    # Rigidity of superstructure
    is_rigid: Bool = True

    # Transformation
    max_allowed_settlement_lim_state_str: confloat(ge=0, le=100000) = 0
    max_allowed_rel_rotation_lim_state_str: confloat(ge=0.0001, le=10000) = 0.0001
    max_allowed_settlement_lim_state_serv: confloat(ge=0, le=100000) = 0
    max_allowed_rel_rotation_lim_state_serv: confloat(ge=0.0001, le=10000) = 0.0001

    # Factors
    factor_xi3: Optional[confloat(ge=0.01, le=10)] = None
    factor_xi4: Optional[confloat(ge=0.01, le=10)] = None
    ea_gem: Optional[confloat(ge=1)] = None

    # Combined Model Options
    is_suppress_qc_reduction: Bool = False
    is_overrule_excavation: Bool = False
    use_pile_group: Bool = True
    is_write_intermediate_results: Bool = False
    use_interaction_model: Bool = False
    use_almere_rules: Bool = False
    use_extra_almere_rules: Bool = False

    def _to_internal(self):
        return InternalModelOptions(**self.dict())

    @classmethod
    def model_type(cls):
        raise NotImplementedError("Implement in concrete classes.")


class BearingPilesModel(ModelOptions):

    factor_gamma_b: Optional[confloat(ge=1, le=100)] = None
    factor_gamma_s: Optional[confloat(ge=1, le=100)] = None
    factor_gamma_fnk: Optional[confloat(ge=-100, le=100)] = None
    area: Optional[confloat(ge=0, le=100000)] = None

    @classmethod
    def model_type(cls):
        return ModelTypeEnum.BEARING_PILES


class TensionPilesModel(ModelOptions):
    unit_weight_water: confloat(ge=0.01, le=20) = 9.81
    use_compaction: Bool = False
    surcharge: confloat(ge=0, le=1e7) = 0
    use_piezometric_levels: Bool = True

    factor_gamma_var: Optional[confloat(ge=0.01, le=100)] = None
    factor_gamma_st: Optional[confloat(ge=0.01, le=100)] = None
    factor_gamma_gamma: Optional[confloat(ge=0.01, le=100)] = None

    @classmethod
    def model_type(cls):
        return ModelTypeEnum.TENSION_PILES


class CalculationOptions(BaseDataClass):
    """
    Set calculation type:

    # Verification
    VERIFICATION_DESIGN = 0  # Only valid for Bearing Piles
    VERIFICATION_COMPLETE = 1

    # Preliminary Design
    INDICATION_BEARING_CAPACITY = 2
    BEARING_CAPACITY_AT_FIXED_PILETIP_LEVELS = 3
    PILETIP_LEVELS_AND_NET_BEARING_CAPACITY = 4

    Note that cpt_test_level is related to the
    chosen CPTs for the calculation. The GUI will
    override the display of an invalid value.
    """

    calculationtype: CalculationType

    net_bearing_capacity: Optional[float] = 0  # [kN]
    cpt_test_level: Optional[float] = 0.0  # [m]

    trajectory_begin: float = -10.00
    trajectory_end: float = -25.00
    trajectory_interval: float = 0.50

    def _to_internal(self):
        fields = self.dict(exclude={"calculationtype"})
        return PreliminaryDesign(**fields)


class DFoundationsModel(BaseModel):
    r"""D-Foundations enables the user to calculate piles (bearing and tension)
    and shallow foundation in accordance with the Dutch (and Belgian) standards.

    This model can read, modify and create \*.foi files, read \*.fod and \*.err files.
    """

    datastructure: Union[
        DFoundationsDumpStructure,
        DFoundationsStructure,
    ] = DFoundationsStructure()

    @property
    def parser_provider_type(self) -> Type[DFoundationsParserProvider]:
        return DFoundationsParserProvider

    @property
    def console_path(self) -> Path:
        return Path("DFoundationsConsole/DFoundationsConsole.exe")

    @property
    def console_flags(self) -> List[str]:
        return [CONSOLE_RUN_BATCH_FLAG]

    @property
    def output(self) -> DFoundationsDumpfileOutputStructure:
        return self.datastructure.dumpfile.dumpfile_output

    @property
    def input(self):
        return self.datastructure.input_data

    def serialize(self, filename: FilePath):
        serializer = DFoundationsInputSerializer(ds=self.datastructure.dict())
        serializer.write(filename)
        self.filename = filename

    def set_model(
        self,
        model: Union[BearingPilesModel, TensionPilesModel],
        calculation: CalculationOptions,
    ) -> None:
        """(Re)Set ModelType (Bearing/Tension) and ConstructionType for model.

        Please note:

        - All profiles will be automatically selected for calculation
        - Only the relevant pile types will be used (bearing/tension):

            - Only the first one added will be used in the case of verification calculation

        - On model change (bearing to tensions and reverse) the soils will be replaced by defaults

        It is advised to only use this method once at the beginning of your workflow.
        """
        logger.warning(
            "Setting model, "
            "prior made modifications in the current D-Foundations model might be overwritten."
        )
        self.datastructure.input_data.model.model = model.model_type()
        self.datastructure.input_data.calculation_options = model._to_internal()
        self.datastructure.input_data.calculationtype = InternalCalculationType(
            sub_calculationtype=calculation.calculationtype
        )
        self.datastructure.input_data.preliminary_design = calculation._to_internal()
        self.datastructure.input_data.preliminary_design.profiles = list(
            range(len(self.profiles.profiles))
        )
        self.datastructure.input_data.soil_collection.soil = InternalSoil.default_soils(
            model=model.model_type().name
        )

    @property
    def soils(self) -> SoilCollection:
        return self.datastructure.input_data.soil_collection

    def add_soil(self, soil: Soil) -> str:
        """Soil is converted in the internal structure and added in soil_collection."""
        soil_internal = soil._to_dfoundations()
        new_soil = self.datastructure.input_data.soil_collection.add_soil_if_unique(
            soil_internal
        )
        return new_soil.name

    @property
    def cpts(self) -> CPTList:
        return self.datastructure.input_data.cpt_list

    def add_profile(self, profile: Profile) -> str:
        """Add profile on location to the model. Requires setup of the Profile and Location first."""

        # Replace materials' names with ids
        for layer in profile.layers:
            soil_id = self.soils.find_soil_id(layer.get("material", None))
            layer["material"] = soil_id

        cpt = profile.cpt._to_internal()

        cpt_id = self.cpts.add_cpt(cpt)
        internal_profile = profile._to_internal(cpt_id)

        profile = self.profiles.add_profile_if_unique(internal_profile)

        # Automatically select all profiles for calculation
        if isinstance(self.datastructure.input_data.preliminary_design, str):
            logger.warning("Overwriting unparsed calculation settings.")
            self.datastructure.input_data.preliminary_design = PreliminaryDesign()
        self.datastructure.input_data.preliminary_design.profiles = list(
            range(len(self.profiles.profiles))
        )

        return profile.name

    def add_pile_if_unique(self, pile: Pile, location: PileLocation):
        """Add pile on location to the model. Requires setup of the Pile and Location first."""
        if isinstance(pile, BearingPile) and isinstance(location, BearingPileLocation):
            piles = self.bearing_piles
            locations = self.bearing_pile_locations
        elif isinstance(pile, TensionPile) and isinstance(location, TensionPileLocation):
            piles = self.tension_piles
            locations = self.tension_pile_locations
        else:
            logger.warning(
                "Pile and Pile Location are mixed tension an bearing pile/location, "
                "the pile and location is not added."
            )
            return

        i_pile = pile._to_internal()
        if i_pile not in piles:
            piles.append(i_pile)

        pile_id = len(locations) + 1
        if location.pile_name == "":
            location.pile_name = "Pos(" + str(pile_id) + ")"
        i_location = location._to_internal(pile_id)
        if i_location not in locations:
            locations.append(i_location)

        # Enable Pile in calculation
        if isinstance(self.datastructure.input_data.preliminary_design, str):
            logger.warning("Overwriting unparsed calculation settings.")
            self.datastructure.input_data.preliminary_design = PreliminaryDesign()
        self.datastructure.input_data.preliminary_design.pile_types = list(
            range(len(piles))
        )

    @property
    def bearing_piles(self):
        """Enables easy access to the bearing piles in the internal dict-like datastructure.
        Also enables edit/delete for individual Piles."""
        if isinstance(self.datastructure.input_data.types___bearing_piles, str):
            self.datastructure.input_data.types___bearing_piles = []

        return self.datastructure.input_data.types___bearing_piles

    @property
    def tension_piles(self):
        """Enables easy access to the tension piles in the internal dict-like datastructure.
        Also enables edit/delete for individual Piles."""
        if isinstance(self.datastructure.input_data.types___tension_piles_cur, str):
            self.datastructure.input_data.types___tension_piles_cur = []

        return self.datastructure.input_data.types___tension_piles_cur

    @property
    def bearing_pile_locations(self):
        """Enables easy access to the bearing piles location in the internal dict-like datastructure.
        Also enables edit/delete for individual bearing pile locations."""
        if isinstance(self.datastructure.input_data.positions___bearing_piles, str):
            self.datastructure.input_data.positions___bearing_piles = (
                PositionsBearingPiles()
            )
        return self.datastructure.input_data.positions___bearing_piles.positions

    @property
    def tension_pile_locations(self):
        """Enables easy access to the tension piles location in the internal dict-like datastructure.
        Also enables edit/delete for individual tension pile locations."""
        if isinstance(self.datastructure.input_data.positions___tension_piles_cur, str):
            self.datastructure.input_data.positions___tension_piles_cur = (
                PositionsTensionPiles()
            )
        return self.datastructure.input_data.positions___tension_piles_cur.positions

    @property
    def profiles(self):
        """Enables easy access to the profiles in the internal dict-like datastructure. Also enables edit/delete for individual Piles."""
        return self.datastructure.input_data.profiles
